from typing import ClassVar, Optional
from unittest import mock

import pytest

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.expectations.metrics.query_metric_provider import (
    QueryMetricProvider,
    QueryParameters,
)
from great_expectations.expectations.metrics.query_metrics import (
    QueryColumn,
    QueryColumnPair,
    QueryMultipleColumns,
    QueryRowCount,
    QueryTable,
    QueryTemplateValues,
)
from tests.expectations.metrics.conftest import MockSqlAlchemyExecutionEngine


@pytest.mark.unit
def test_query_template_get_query_function_with_int():
    """Simple test to ensure that the `get_query()` method for QueryTemplateValue can handle integer value"""  # noqa: E501 # FIXME CoP
    query: str = """
            SELECT {column_to_check}
            FROM {batch}
            WHERE {condition}
            GROUP BY {column_to_check}
            """
    selectable = sa.Table("gx_temp_aaa", sa.MetaData(), schema=None)
    template_dict: dict = {"column_to_check": 1, "condition": "is_open"}
    metric_ob: QueryTemplateValues = QueryTemplateValues()
    formatted_query: str = metric_ob.get_query(query, template_dict, selectable)
    assert (
        formatted_query
        == """
            SELECT 1
            FROM gx_temp_aaa
            WHERE is_open
            GROUP BY 1
            """
    )


@pytest.mark.unit
def test_query_template_get_query_function_with_float():
    """Simple test to ensure that the `get_query()` method for QueryTemplateValue can handle float value"""  # noqa: E501 # FIXME CoP
    query: str = """
            SELECT {column_to_check}
            FROM {batch}
            WHERE {condition}
            GROUP BY {column_to_check}
            """
    selectable = sa.Table("gx_temp_aaa", sa.MetaData(), schema=None)
    template_dict: dict = {"column_to_check": 1.0, "condition": "is_open"}
    metric_ob: QueryTemplateValues = QueryTemplateValues()
    formatted_query: str = metric_ob.get_query(query, template_dict, selectable)
    assert (
        formatted_query
        == """
            SELECT 1.0
            FROM gx_temp_aaa
            WHERE is_open
            GROUP BY 1.0
            """
    )


class MyQueryColumn(QueryColumn):
    metric_name = "my_query.column"
    value_keys = ("my_query",)

    query_param_name: ClassVar[str] = "my_query"


class MyQueryColumnPair(QueryColumnPair):
    metric_name = "my_query.column_pair"
    value_keys = ("my_query",)

    query_param_name: ClassVar[str] = "my_query"


class MyQueryMultipleColumns(QueryMultipleColumns):
    metric_name = "my_query.multiple_columns"
    value_keys = ("my_query",)

    query_param_name: ClassVar[str] = "my_query"


class MyQueryTable(QueryTable):
    metric_name = "my_query.table"
    value_keys = ("my_query",)

    query_param_name: ClassVar[str] = "my_query"


@pytest.mark.unit
@mock.patch.object(sa, "text")
@mock.patch.object(
    QueryMetricProvider, "_get_substituted_batch_subquery_from_query_and_batch_selectable"
)
@mock.patch.object(QueryMetricProvider, "_get_sqlalchemy_records_from_substituted_batch_subquery")
@pytest.mark.parametrize(
    "metric_class, class_metric_value_kwargs, query_parameters",
    [
        (
            MyQueryColumn,
            {"column": "my_column"},
            QueryParameters(column="my_column"),
        ),
        (
            MyQueryColumnPair,
            {"column_A": "my_column_A", "column_B": "my_column_B"},
            QueryParameters(column_A="my_column_A", column_B="my_column_B"),
        ),
        (
            MyQueryMultipleColumns,
            {"columns": ["my_column_1", "my_column_2", "my_column_3"]},
            QueryParameters(columns=["my_column_1", "my_column_2", "my_column_3"]),
        ),
        (
            MyQueryTable,
            {},
            None,
        ),
    ],
)
def test_sqlalchemy_query_metrics_that_return_records(
    mock_get_sqlalchemy_records_from_substituted_batch_subquery,
    mock_get_substituted_batch_subquery_from_query_and_batch_selectable,
    mock_sqlalchemy_text,
    mock_sqlalchemy_execution_engine: MockSqlAlchemyExecutionEngine,
    metric_class: QueryMetricProvider,
    class_metric_value_kwargs: dict,
    query_parameters: Optional[QueryParameters],
    batch_selectable: sa.Table,
):
    metric_value_kwargs = {
        "query_param": "my_query",
        "my_query": "SELECT * FROM {batch} WHERE passenger_count > 7",
    }
    metric_value_kwargs.update(class_metric_value_kwargs)

    mock_substituted_batch_subquery = "SELECT * FROM (my_table) WHERE passenger_count > 7"
    mock_get_substituted_batch_subquery_from_query_and_batch_selectable.return_value = (
        mock_substituted_batch_subquery
    )
    mock_sqlalchemy_text.return_value = "*"
    with mock.patch.object(mock_sqlalchemy_execution_engine, "execute_query"):
        metric_class._sqlalchemy(
            cls=metric_class,
            execution_engine=mock_sqlalchemy_execution_engine,
            metric_domain_kwargs={},
            metric_value_kwargs=metric_value_kwargs,
            metrics={},
            runtime_configuration={},
        )
    if query_parameters:
        mock_get_substituted_batch_subquery_from_query_and_batch_selectable.assert_called_once_with(
            query=metric_value_kwargs["my_query"],
            batch_selectable=batch_selectable,
            execution_engine=mock_sqlalchemy_execution_engine,
            query_parameters=query_parameters,
        )
    else:
        mock_get_substituted_batch_subquery_from_query_and_batch_selectable.assert_called_once_with(
            query=metric_value_kwargs["my_query"],
            batch_selectable=batch_selectable,
            execution_engine=mock_sqlalchemy_execution_engine,
        )
    mock_get_sqlalchemy_records_from_substituted_batch_subquery.assert_called_once_with(
        substituted_batch_subquery=mock_substituted_batch_subquery,
        execution_engine=mock_sqlalchemy_execution_engine,
    )


class MyQueryRowCount(QueryRowCount):
    metric_name = "my_query.row_count"
    value_keys = ("my_query",)

    query_param_name: ClassVar[str] = "my_query"


@pytest.mark.unit
@mock.patch.object(sa, "text")
@mock.patch.object(
    QueryMetricProvider, "_get_substituted_batch_subquery_from_query_and_batch_selectable"
)
def test_sqlalchemy_query_row_count(
    mock_get_substituted_batch_subquery_from_query_and_batch_selectable,
    mock_sqlalchemy_text,
    mock_sqlalchemy_execution_engine: MockSqlAlchemyExecutionEngine,
    batch_selectable: sa.Table,
):
    metric_value_kwargs = {
        "query_param": "my_query",
        "my_query": "SELECT * FROM {batch} WHERE passenger_count > 7",
    }

    mock_substituted_batch_subquery = "SELECT * FROM (my_table) WHERE passenger_count > 7"
    mock_get_substituted_batch_subquery_from_query_and_batch_selectable.return_value = (
        mock_substituted_batch_subquery
    )
    mock_sqlalchemy_text.return_value = "*"
    with mock.patch.object(mock_sqlalchemy_execution_engine, "execute_query"):
        MyQueryRowCount._sqlalchemy(
            cls=MyQueryRowCount,
            execution_engine=mock_sqlalchemy_execution_engine,
            metric_domain_kwargs={},
            metric_value_kwargs=metric_value_kwargs,
            metrics={},
            runtime_configuration={},
        )
    mock_get_substituted_batch_subquery_from_query_and_batch_selectable.assert_called_once_with(
        query=metric_value_kwargs["my_query"],
        batch_selectable=batch_selectable,
        execution_engine=mock_sqlalchemy_execution_engine,
    )



import pytest

from great_expectations.expectations.row_conditions import (
    _parse_great_expectations_condition,
    parse_condition_to_spark,
    parse_condition_to_sqlalchemy,
)


@pytest.mark.unit
def test_notnull_parser():
    res = _parse_great_expectations_condition('col("foo").notNull()')
    assert res["column"] == "foo"
    assert res["notnull"] is True

    res = _parse_great_expectations_condition('col("foo") > 5')
    assert res["column"] == "foo"
    assert "notnull" not in res


@pytest.mark.unit
def test_condition_parser():
    res = _parse_great_expectations_condition('col("foo") > 5')
    assert res["column"] == "foo"
    assert res["op"] == ">"
    assert res["fnumber"] == "5"


@pytest.mark.unit
def test_condition_parser_with_underscore_in_column_name():
    res = _parse_great_expectations_condition('col("pk_2") == "Two"')
    assert res["column"] == "pk_2"
    assert res["op"] == "=="
    assert res["condition_value"] == "Two"


@pytest.mark.unit
def test_condition_parser_with_dash_in_column_name():
    res = _parse_great_expectations_condition('col("pk-2") == "Two"')
    assert res["column"] == "pk-2"
    assert res["op"] == "=="
    assert res["condition_value"] == "Two"


@pytest.mark.unit
def test_condition_parser_with_space_in_column_name():
    res = _parse_great_expectations_condition('col("pk 2") == "Two"')
    assert res["column"] == "pk 2"
    assert res["op"] == "=="
    assert res["condition_value"] == "Two"


@pytest.mark.unit
def test_condition_parser_with_space_in_condition_value():
    res = _parse_great_expectations_condition('col("pk_2") == "Two Two"')
    assert res["column"] == "pk_2"
    assert res["op"] == "=="
    assert res["condition_value"] == "Two Two"


@pytest.mark.unit
def test_condition_parser_with_tab_in_condition_value():
    res = _parse_great_expectations_condition('col("pk_2") == "Two  Two"')
    assert res["column"] == "pk_2"
    assert res["op"] == "=="
    assert res["condition_value"] == "Two  Two"


@pytest.mark.unit
def test_condition_parser_with_8bit_chars_in_condition_value():
    res = _parse_great_expectations_condition(
        'col("sp_chars") == "Café_über_naïve_élite_ñandú_çeviche_æther_øresund_ångström"'
    )
    assert res["column"] == "sp_chars"
    assert res["op"] == "=="
    assert res["condition_value"] == "Café_über_naïve_élite_ñandú_çeviche_æther_øresund_ångström"


@pytest.mark.spark
def test_parse_condition_to_spark(spark_session):
    res = parse_condition_to_spark('col("foo") > 5')
    # This is mostly a demonstrative test; it may be brittle. I do not know how to test
    # a condition itself.
    assert str(res) in ["Column<b'(foo > 5)'>", "Column<'(foo > 5)'>"]

    res = parse_condition_to_spark('col("foo").notNull()')
    # This is mostly a demonstrative test; it may be brittle. I do not know how to test
    # a condition itself.
    print(str(res))
    assert str(res) in ["Column<b'(foo IS NOT NULL)'>", "Column<'(foo IS NOT NULL)'>"]


@pytest.mark.unit
def test_parse_condition_to_sqlalchemy(sa):
    res = parse_condition_to_sqlalchemy('col("foo") > 5')
    assert str(res) == "foo > :foo_1"

    res = parse_condition_to_sqlalchemy('col("foo").notNull()')
    assert str(res) == "foo IS NOT NULL"

    res = parse_condition_to_sqlalchemy('col("foo") == "a-b"')
    assert str(res) == "foo = :foo_1"

    res = parse_condition_to_sqlalchemy('col("foo") != "a-b"')
    assert str(res) == "foo != :foo_1"

    res = parse_condition_to_sqlalchemy('col("foo") <= 1.34')
    assert str(res) == "foo <= :foo_1"

    res = parse_condition_to_sqlalchemy('col("foo") <= date("2023-03-13")')
    assert str(res) == "foo <= :foo_1"



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from great_expectations.core.expectation_validation_result import (
        ExpectationValidationResult,
    )


class ValidationStatistics(NamedTuple):
    evaluated_expectations: int
    successful_expectations: int
    unsuccessful_expectations: int
    success_percent: float | None
    success: bool


def calc_validation_statistics(
    validation_results: list[ExpectationValidationResult],
) -> ValidationStatistics:
    """
    Calculate summary statistics for the validation results and
    return ``ExpectationStatistics``.
    """
    # calc stats
    evaluated_expectations = len(validation_results)
    successful_expectations = len([exp for exp in validation_results if exp.success])
    unsuccessful_expectations = evaluated_expectations - successful_expectations
    success = successful_expectations == evaluated_expectations
    try:
        success_percent = successful_expectations / evaluated_expectations * 100
    except ZeroDivisionError:
        success_percent = None

    return ValidationStatistics(
        successful_expectations=successful_expectations,
        evaluated_expectations=evaluated_expectations,
        unsuccessful_expectations=unsuccessful_expectations,
        success=success,
        success_percent=success_percent,
    )



from collections.abc import Sequence

import pandas as pd
import pytest
from sqlalchemy import types as sqlatypes

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import RedshiftDatasourceTestConfig
from tests.integration.test_utils.data_source_config.base import DataSourceTestConfig
from tests.integration.test_utils.data_source_config.big_query import BigQueryDatasourceTestConfig
from tests.integration.test_utils.data_source_config.databricks import (
    DatabricksDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.mssql import MSSQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.mysql import MySQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.pandas_data_frame import (
    PandasDataFrameDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.pandas_filesystem_csv import (
    PandasFilesystemCsvDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.postgres import PostgreSQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.snowflake import SnowflakeDatasourceTestConfig
from tests.integration.test_utils.data_source_config.spark_filesystem_csv import (
    SparkFilesystemCsvDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.sqlite import SqliteDatasourceTestConfig

ALL_UNIQUE_COL = "all_unique"
NO_UNIQUE_COL = "one_unique"
SOME_UNIQUE_COL = "some_unique"
STRING_COL = "my_strings"


A_LOT_OF_NONES = [None] * 10
DATA = pd.DataFrame(
    {
        ALL_UNIQUE_COL: [0, 1, 2, 3, *A_LOT_OF_NONES],
        NO_UNIQUE_COL: [None, None, None, None, *A_LOT_OF_NONES],
        SOME_UNIQUE_COL: [1, 2, 3, 3, *A_LOT_OF_NONES],
        STRING_COL: ["foo", "foo", "bar", "baz", *A_LOT_OF_NONES],
    },
    dtype="object",
)

COLUMN_TYPES = {NO_UNIQUE_COL: sqlatypes.INTEGER}
try:
    from great_expectations.compatibility.pyspark import types as PYSPARK_TYPES

    SPARK_COLUMN_TYPES = {
        ALL_UNIQUE_COL: PYSPARK_TYPES.IntegerType,
        NO_UNIQUE_COL: PYSPARK_TYPES.IntegerType,
        SOME_UNIQUE_COL: PYSPARK_TYPES.IntegerType,
        STRING_COL: PYSPARK_TYPES.StringType,
    }
except ModuleNotFoundError:
    SPARK_COLUMN_TYPES = {}

ALL_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(column_types=COLUMN_TYPES),
    DatabricksDatasourceTestConfig(column_types=COLUMN_TYPES),
    MSSQLDatasourceTestConfig(column_types=COLUMN_TYPES),
    MySQLDatasourceTestConfig(column_types=COLUMN_TYPES),
    PandasDataFrameDatasourceTestConfig(),
    PandasFilesystemCsvDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(column_types=COLUMN_TYPES),
    RedshiftDatasourceTestConfig(column_types=COLUMN_TYPES),
    SnowflakeDatasourceTestConfig(column_types=COLUMN_TYPES),
    SparkFilesystemCsvDatasourceTestConfig(column_types=SPARK_COLUMN_TYPES),
    SqliteDatasourceTestConfig(column_types=COLUMN_TYPES),
]


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
        column=SOME_UNIQUE_COL, min_value=0.75, max_value=0.75
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {"observed_value": 0.75}


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_strings(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
        column=STRING_COL, min_value=0.75, max_value=0.75
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(column=SOME_UNIQUE_COL),
            id="vacuously_true",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL, min_value=0.75, max_value=0.75
            ),
            id="some_unique",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=ALL_UNIQUE_COL, min_value=1, max_value=1
            ),
            id="all_unique",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=NO_UNIQUE_COL, min_value=0, max_value=0
            ),
            id="no_unique",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL, min_value=0.75
            ),
            id="only_min",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL, min_value=0.75
            ),
            id="only_max",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL,
                min_value=0.74,
                max_value=0.76,
                strict_min=True,
                strict_max=True,
            ),
            id="strict_bounds",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnProportionOfUniqueValuesToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL, min_value=0.75, strict_min=True
            ),
            id="strict_min",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL, max_value=0.75, strict_max=True
            ),
            id="strict_max",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
                column=SOME_UNIQUE_COL, min_value=0.8, max_value=0.9
            ),
            id="wrong_bounds",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnProportionOfUniqueValuesToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(False, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_strict_min_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_proportion_of_unique_values_to_be_between"
    expectation = gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
        column=SOME_UNIQUE_COL,
        min_value=0.75,
        max_value=0.75,
        strict_min={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(False, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_strict_max_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_proportion_of_unique_values_to_be_between"
    expectation = gxe.ExpectColumnProportionOfUniqueValuesToBeBetween(
        column=SOME_UNIQUE_COL,
        min_value=0.75,
        max_value=0.75,
        strict_max={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



import datetime
import uuid

import boto3
import pytest
from freezegun import freeze_time
from moto import mock_s3

from great_expectations.core import ExpectationSuiteValidationResult
from great_expectations.data_context.store import ValidationResultsStore
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.util import gen_directory_tree_str
from tests import test_utils


@freeze_time("09/26/2019 13:42:41")
@mock_s3
@pytest.mark.filterwarnings(
    "ignore:String run_ids are deprecated*:DeprecationWarning:great_expectations.data_context.types.resource_identifiers"  # noqa: E501 # FIXME CoP
)
@pytest.mark.aws_deps
def test_ValidationResultsStore_with_TupleS3StoreBackend(aws_credentials):
    bucket = "test_validation_store_bucket"
    prefix = "test/prefix"

    # create a bucket in Moto's mock AWS environment
    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    # First, demonstrate that we pick up default configuration including from an S3TupleS3StoreBackend  # noqa: E501 # FIXME CoP
    my_store = ValidationResultsStore(
        store_backend={
            "class_name": "TupleS3StoreBackend",
            "bucket": bucket,
            "prefix": prefix,
        }
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    ns_1 = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier(
            name="asset.quarantine",
        ),
        run_id="20191007T151224.1234Z_prod_100",
        batch_identifier="batch_id",
    )
    my_store.set(
        ns_1,
        ExpectationSuiteValidationResult(success=True, results=[], suite_name="asset.quarantine"),
    )
    assert my_store.get(ns_1) == ExpectationSuiteValidationResult(
        success=True, statistics={}, results=[], suite_name="asset.quarantine"
    )

    ns_2 = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier(
            name="asset.quarantine",
        ),
        run_id="20191007T151224.1234Z_prod_200",
        batch_identifier="batch_id",
    )

    my_store.set(
        ns_2,
        ExpectationSuiteValidationResult(success=False, results=[], suite_name="asset.quarantine"),
    )
    assert my_store.get(ns_2) == ExpectationSuiteValidationResult(
        success=False, statistics={}, results=[], suite_name="asset.quarantine"
    )

    # Verify that internals are working as expected, including the default filepath
    assert {
        s3_object_info["Key"]
        for s3_object_info in boto3.client("s3").list_objects_v2(Bucket=bucket, Prefix=prefix)[
            "Contents"
        ]
    } == {
        "test/prefix/.ge_store_backend_id",
        "test/prefix/asset/quarantine/20191007T151224.1234Z_prod_100/20190926T134241.000000Z/batch_id.json",
        "test/prefix/asset/quarantine/20191007T151224.1234Z_prod_200/20190926T134241.000000Z/batch_id.json",
    }

    print(my_store.list_keys())
    assert set(my_store.list_keys()) == {
        ns_1,
        ns_2,
    }

    """
    What does this test and why?
    A Store should be able to report it's store_backend_id
    which is set when the StoreBackend is instantiated.
    """
    # Check that store_backend_id exists can be read
    assert my_store.store_backend_id is not None
    # Check that store_backend_id is a valid UUID
    assert isinstance(my_store.store_backend_id, uuid.UUID)


@freeze_time("09/26/2019 13:42:41")
@pytest.mark.big
def test_ValidationResultsStore_with_InMemoryStoreBackend():
    my_store = ValidationResultsStore(
        store_backend={
            "module_name": "great_expectations.data_context.store",
            "class_name": "InMemoryStoreBackend",
        }
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    ns_1 = ValidationResultIdentifier.from_tuple(
        (
            "a",
            "b",
            "c",
            "quarantine",
            datetime.datetime.now(datetime.timezone.utc),
            "prod-100",
        )
    )
    my_store.set(
        ns_1,
        ExpectationSuiteValidationResult(success=True, results=[], suite_name="a.b.c.quarantine"),
    )
    assert my_store.get(ns_1) == ExpectationSuiteValidationResult(
        success=True, statistics={}, results=[], suite_name="a.b.c.quarantine"
    )

    ns_2 = ValidationResultIdentifier.from_tuple(
        (
            "a",
            "b",
            "c",
            "quarantine",
            datetime.datetime.now(datetime.timezone.utc),
            "prod-200",
        )
    )
    my_store.set(
        ns_2,
        ExpectationSuiteValidationResult(success=False, results=[], suite_name="a.b.c.quarantine"),
    )
    assert my_store.get(ns_2) == ExpectationSuiteValidationResult(
        success=False, statistics={}, results=[], suite_name="a.b.c.quarantine"
    )
    assert set(my_store.list_keys()) == {
        ns_1,
        ns_2,
    }
    """
    What does this test and why?
    A Store should be able to report it's store_backend_id
    which is set when the StoreBackend is instantiated.
    """
    # Check that store_backend_id exists can be read
    assert my_store.store_backend_id is not None
    # Check that store_backend_id is a valid UUID
    assert isinstance(my_store.store_backend_id, uuid.UUID)


@pytest.mark.big
@freeze_time("09/26/2019 13:42:41")
@pytest.mark.filterwarnings(
    "ignore:String run_ids are deprecated*:DeprecationWarning:great_expectations.data_context.types.resource_identifiers"  # noqa: E501 # FIXME CoP
)
def test_ValidationResultsStore_with_TupleFileSystemStoreBackend(tmp_path_factory):
    full_test_dir = tmp_path_factory.mktemp(
        "test_ValidationResultStore_with_TupleFileSystemStoreBackend__dir"
    )
    test_dir = full_test_dir.parts[-1]
    path = str(full_test_dir)

    my_store = ValidationResultsStore(
        store_backend={
            "module_name": "great_expectations.data_context.store",
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": "my_store/",
        },
        runtime_environment={"root_directory": path},
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    ns_1 = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier("asset.quarantine"),
        run_id="prod-100",
        batch_identifier="batch_id",
    )
    my_store.set(
        ns_1,
        ExpectationSuiteValidationResult(success=True, results=[], suite_name="asset.quarantine"),
    )
    assert my_store.get(ns_1) == ExpectationSuiteValidationResult(
        success=True, statistics={}, results=[], suite_name="asset.quarantine"
    )

    ns_2 = ValidationResultIdentifier.from_tuple(
        (
            "asset",
            "quarantine",
            "prod-20",
            datetime.datetime.now(datetime.timezone.utc),
            "batch_id",
        )
    )
    my_store.set(
        ns_2,
        ExpectationSuiteValidationResult(success=False, results=[], suite_name="asset.quarantine"),
    )
    assert my_store.get(ns_2) == ExpectationSuiteValidationResult(
        success=False, statistics={}, results=[], suite_name="asset.quarantine"
    )

    assert set(my_store.list_keys()) == {
        ns_1,
        ns_2,
    }

    assert (
        gen_directory_tree_str(path)
        == f"""\
{test_dir}/
    my_store/
        .ge_store_backend_id
        asset/
            quarantine/
                prod-100/
                    20190926T134241.000000Z/
                        batch_id.json
                prod-20/
                    20190926T134241.000000Z/
                        batch_id.json
"""
    )

    """
    What does this test and why?
    A Store should be able to report it's store_backend_id
    which is set when the StoreBackend is instantiated.
    """
    # Check that store_backend_id exists can be read
    assert my_store.store_backend_id is not None
    # Check that store_backend_id is a valid UUID
    assert isinstance(my_store.store_backend_id, uuid.UUID)

    # Check that another store with the same configuration shares the same store_backend_id
    my_store_duplicate = ValidationResultsStore(
        store_backend={
            "module_name": "great_expectations.data_context.store",
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": "my_store/",
        },
        runtime_environment={"root_directory": path},
    )
    assert my_store.store_backend_id == my_store_duplicate.store_backend_id


@pytest.mark.filterwarnings(
    "ignore:String run_ids are deprecated*:DeprecationWarning:great_expectations.data_context.types.resource_identifiers"  # noqa: E501 # FIXME CoP
)
@pytest.mark.big
def test_ValidationResultsStore_with_DatabaseStoreBackend(sa):
    # Use sqlite so we don't require postgres for this test.
    connection_kwargs = {"drivername": "sqlite"}

    # First, demonstrate that we pick up default configuration
    my_store = ValidationResultsStore(
        store_backend={
            "class_name": "DatabaseStoreBackend",
            "credentials": connection_kwargs,
        }
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    ns_1 = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier(
            name="asset.quarantine",
        ),
        run_id="20191007T151224.1234Z_prod_100",
        batch_identifier="batch_id",
    )
    my_store.set(
        ns_1,
        ExpectationSuiteValidationResult(success=True, results=[], suite_name="asset.quarantine"),
    )
    assert my_store.get(ns_1) == ExpectationSuiteValidationResult(
        success=True, statistics={}, results=[], suite_name="asset.quarantine"
    )

    ns_2 = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier(
            name="asset.quarantine",
        ),
        run_id="20191007T151224.1234Z_prod_200",
        batch_identifier="batch_id",
    )

    my_store.set(
        ns_2,
        ExpectationSuiteValidationResult(success=False, results=[], suite_name="asset.quarantine"),
    )
    assert my_store.get(ns_2) == ExpectationSuiteValidationResult(
        success=False, statistics={}, results=[], suite_name="asset.quarantine"
    )

    assert set(my_store.list_keys()) == {
        ns_1,
        ns_2,
    }

    """
    What does this test and why?
    A Store should be able to report it's store_backend_id
    which is set when the StoreBackend is instantiated.
    """
    # Check that store_backend_id exists can be read
    assert my_store.store_backend_id is not None
    # Check that store_backend_id is a valid UUID
    assert test_utils.validate_uuid4(my_store.store_backend_id)


@pytest.mark.cloud
def test_gx_cloud_response_json_to_object_dict() -> None:
    validation_id = "c1e8f964-ba44-4a13-a9b6-7331a358f12d"
    validation_definition = {
        "results": [],
        "success": True,
        "statistics": {
            "evaluated_expectations": 0,
            "successful_expectations": 0,
            "unsuccessful_expectations": 0,
            "success_percent": None,
        },
    }
    response_json = {
        "data": {
            "id": validation_id,
            "attributes": {
                "result": validation_definition,
            },
        }
    }

    expected = validation_definition
    expected["id"] = validation_id

    actual = ValidationResultsStore.gx_cloud_response_json_to_object_dict(response_json)

    assert actual == expected



from __future__ import annotations

from typing import Optional, Union

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)


class _SparkGenericFilePathAssetMixin(PathDataAsset):
    # vvv Docs <> Source Code mismatch
    # ignoreCorruptFiles and ignoreMissingFiles appear in the docs https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html
    # but not in any reader method signatures (e.g. https://github.com/apache/spark/blob/v3.4.0/python/pyspark/sql/readwriter.py#L604)
    # ignore_corrupt_files: bool = Field(alias="ignoreCorruptFiles")
    # ignore_missing_files: bool = Field(alias="ignoreMissingFiles")
    # ^^^ Docs <> Source Code mismatch

    path_glob_filter: Optional[Union[bool, str]] = Field(None, alias="pathGlobFilter")
    recursive_file_lookup: Optional[Union[bool, str]] = Field(None, alias="recursiveFileLookup")
    modified_before: Optional[Union[bool, str]] = Field(None, alias="modifiedBefore")
    modified_after: Optional[Union[bool, str]] = Field(None, alias="modifiedAfter")

    @override
    def _get_reader_options_include(self) -> set[str]:
        return {
            "path_glob_filter",
            "recursive_file_lookup",
            "modified_before",
            "modified_after",
            # vvv Missing from method signatures but appear in documentation:
            # "ignoreCorruptFiles",
            # "ignore_missing_files",
            # ^^^ Missing from method signatures but appear in documentation:
        }



import boto3

from great_expectations.compatibility import pyspark
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import get_context
from great_expectations.data_context.types.base import (
    DataContextConfig,
    S3StoreBackendDefaults,
)
from great_expectations.execution_engine import SparkDFExecutionEngine

yaml = YAMLHandler()
# </snippet>

if __name__ == "__main__":
    ### critical part to reinitialize spark context
    sc = SparkDFExecutionEngine.get_or_create_spark_session()
    spark = pyspark.SQLContext(sc)

    spark_file = "pyspark_df.parquet"
    suite_name = "pandas_spark_suite"
    session = boto3.Session()
    s3_client = session.client("s3")
    response = s3_client.get_object(
        Bucket="bucket_name",
        Key="bucket_name/great_expectations/great_expectations.yml",
    )
    config_file = yaml.load(response["Body"])
    df_spark = spark.read.parquet("s3://bucket_name/data_folder/" + spark_file)

    config = DataContextConfig(
        config_version=config_file["config_version"],
        datasources=config_file["datasources"],
        expectations_store_name=config_file["expectations_store_name"],
        validation_results_store_name=config_file["validation_results_store_name"],
        plugins_directory="/great_expectations/plugins",
        stores=config_file["stores"],
        data_docs_sites=config_file["data_docs_sites"],
        config_variables_file_path=config_file["config_variables_file_path"],
        anonymous_usage_statistics=config_file["anonymous_usage_statistics"],
        checkpoint_store_name=config_file["checkpoint_store_name"],
        store_backend_defaults=S3StoreBackendDefaults(
            default_bucket_name=config_file["data_docs_sites"]["s3_site"]["store_backend"]["bucket"]
        ),
    )

    context_gx = get_context(project_config=config)
    expectation_suite_name = suite_name
    suite = context_gx.suites.get(suite_name)

    batch_request = RuntimeBatchRequest(
        datasource_name="spark_s3",
        data_connector_name="default_inferred_data_connector_name",
        data_asset_name="datafile_name",
        batch_identifiers={"runtime_batch_identifier_name": "default_identifier"},
        runtime_parameters={"path": "s3a://bucket_name/path_to_file.format"},
    )
    validator = context_gx.get_validator(
        batch_request=batch_request,
        expectation_suite_name=expectation_suite_name,
    )
    print(validator.head())
    validator.expect_column_values_to_not_be_null(column="passenger_count")  ## add some test
    validator.save_expectation_suite(discard_failed_expectations=False)
    my_checkpoint_name = "in_memory_checkpoint"
    python_config = {
        "name": my_checkpoint_name,
        "action_list": [
            {
                "name": "store_validation_result",
                "action": {"class_name": "StoreValidationResultAction"},
            },
        ],
        "validations": [
            {
                "batch_request": {
                    "datasource_name": "spark_s3",
                    "data_connector_name": "default_runtime_data_connector_name",
                    "data_asset_name": "pyspark_df",
                },
                "expectation_suite_name": expectation_suite_name,
            }
        ],
    }
    checkpoint = context_gx.add_or_update_checkpoint(**python_config)

    results = checkpoint.run(
        run_name="run_name",
        batch_request={
            "runtime_parameters": {"batch_data": df_spark},
            "batch_identifiers": {"runtime_batch_identifier_name": "default_identifier"},
        },
    )

    validation_result_identifier = results.list_validation_result_identifiers()[0]
    context_gx.build_data_docs()



from __future__ import annotations

import pytest

from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)


@pytest.fixture
def config1():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def config2():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def config3():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is another expectation."},
    )


@pytest.fixture
def config4():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "COMPLETE"},
        meta={"notes": "This is another expectation."},
    )


@pytest.fixture
def config5():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "a",
            "value_set": [1, 2],  # differs from others
            "result_format": "COMPLETE",
        },
        meta={"notes": "This is another expectation."},
    )


@pytest.fixture
def config6():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "a",
            "value_set": [1, 2, 3, 4],  # differs from others
            "result_format": "COMPLETE",
        },
        meta={"notes": "This is another expectation."},
    )


@pytest.fixture
def config7():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "a",
            "value_set": [1, 2, 3, 4],
        },  # differs from others
        meta={"notes": "This is another expectation."},
    )


@pytest.fixture
def config8():
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        description="The values should be in the specified set",
        kwargs={
            "column": "a",
            "value_set": [1, 2, 3, 4],
        },
        meta={"notes": "This is another expectation."},
    )


@pytest.mark.unit
def test_expectation_configuration_equality(config1, config2, config3, config4):
    """Equality should depend on all defined properties of a configuration object, but not on whether the *instances*
    are the same."""  # noqa: E501 # FIXME CoP
    assert config1 is config1  # no difference  # noqa: PLR0124 # FIXME CoP
    assert config1 is not config2  # different instances, but same content
    assert config1 == config2  # different instances, but same content
    assert not (config1 != config2)  # ne works properly
    assert config1 != config3  # different meta
    assert config1 != config3  # ne works properly
    assert config3 != config4  # different result format


@pytest.mark.unit
def test_expectation_configuration_equivalence(config1, config2, config3, config4, config5):
    """Equivalence should depend only on properties that affect the result of the expectation."""
    assert config1.isEquivalentTo(config2, match_type="runtime")  # no difference
    assert config2.isEquivalentTo(config1, match_type="runtime")
    assert config1.isEquivalentTo(config3, match_type="runtime")  # different meta
    assert config1.isEquivalentTo(config4, match_type="success")  # different result format
    assert not config1.isEquivalentTo(config5, match_type="success")  # different value_set
    assert config1.isEquivalentTo(config5, match_type="domain")  # different result format


@pytest.mark.parametrize(
    "notes",
    [
        pytest.param("my notes", id="string notes"),
        pytest.param(["my", "list", "of", "notes"], id="string notes"),
        pytest.param(None, id="no notes"),
    ],
)
@pytest.mark.unit
def test_expectation_configuration_to_domain_obj(notes: str | list[str] | None):
    expectation_type = "expect_column_values_to_be_in_set"
    column = "genre_id"
    input_set = {1, 2, 3}
    expected_list = list(input_set)
    meta = {"foo": "bar"}

    config = ExpectationConfiguration(
        type=expectation_type,
        kwargs={"column": column, "value_set": input_set},
        notes=notes,
        meta=meta,
    )
    expectation = config.to_domain_obj()

    # Check that the expectation object has the same properties as the config
    assert expectation.expectation_type == expectation_type
    assert expectation.column == column
    assert sorted(expectation.value_set) == sorted(expected_list)
    assert expectation.notes == notes
    assert expectation.meta == meta

    # Ensure that translation to/from config is consistent
    assert expectation.configuration == config


@pytest.mark.unit
def test_expectation_configuration_to_json_dict(config1, config4, config8):
    assert config1.to_json_dict() == {
        "kwargs": {"column": "a", "result_format": "BASIC", "value_set": [1, 2, 3]},
        "meta": {"notes": "This is an expectation."},
        "type": "expect_column_values_to_be_in_set",
    }
    assert config4.to_json_dict() == {
        "kwargs": {"column": "a", "result_format": "COMPLETE", "value_set": [1, 2, 3]},
        "meta": {"notes": "This is another expectation."},
        "type": "expect_column_values_to_be_in_set",
    }
    assert config8.to_json_dict() == {
        "description": "The values should be in the specified set",
        "kwargs": {"column": "a", "value_set": [1, 2, 3, 4]},
        "meta": {"notes": "This is another expectation."},
        "type": "expect_column_values_to_be_in_set",
    }


class TestExpectationConfigurationHash:
    @pytest.mark.unit
    def test_hash_consistency_with_equality(self, config1, config2):
        assert config1 == config2
        assert hash(config1) == hash(config2)

    @pytest.mark.unit
    def test_hash_different_for_different_types(self):
        config1 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )
        config2 = ExpectationConfiguration(
            type="expect_column_values_to_be_between", kwargs={"column": "test_column"}
        )

        assert config1 != config2
        assert hash(config1) != hash(config2)

    @pytest.mark.unit
    def test_hash_different_for_different_kwargs(self):
        config1 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column_1"}
        )
        config2 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column_2"}
        )

        assert config1 != config2
        assert hash(config1) != hash(config2)

    @pytest.mark.unit
    def test_hash_different_for_different_meta(self):
        config1 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null",
            kwargs={"column": "test_column"},
            meta={"test": "value1"},
        )
        config2 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null",
            kwargs={"column": "test_column"},
            meta={"test": "value2"},
        )

        assert config1 != config2
        assert hash(config1) != hash(config2)

    @pytest.mark.unit
    def test_hash_stable_across_runs(self, config1):
        hash1 = hash(config1)
        hash2 = hash(config1)
        hash3 = hash(config1)

        assert hash1 == hash2 == hash3



import pytest

from great_expectations.core import Domain
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.validator.metric_configuration import MetricConfiguration


@pytest.mark.unit
def test_metric_configuration__repr__and__str__(
    table_head_metric_config: MetricConfiguration,
) -> None:
    expected = '{\n  "metric_name": "table.head",\n  "metric_domain_kwargs": {\n    "batch_id": "abc123"\n  },\n  "metric_domain_kwargs_id": "batch_id=abc123",\n  "metric_value_kwargs": {\n    "n_rows": 5\n  },\n  "metric_value_kwargs_id": "n_rows=5",\n  "id": [\n    "table.head",\n    "batch_id=abc123",\n    "n_rows=5"\n  ]\n}'  # noqa: E501 # FIXME CoP
    assert table_head_metric_config.__repr__() == expected
    assert table_head_metric_config.__str__() == expected


@pytest.mark.unit
def test_metric_configuration_metric_dependencies_property_and_setter(
    table_head_metric_config: MetricConfiguration,
) -> None:
    assert table_head_metric_config.metric_dependencies == {}

    table_head_metric_config.metric_dependencies = {"e": "f"}

    assert table_head_metric_config.metric_dependencies == {"e": "f"}


@pytest.mark.unit
def test_metric_configuration_to_json_dict(
    table_head_metric_config: MetricConfiguration,
) -> None:
    assert table_head_metric_config.to_json_dict() == {
        "id": ["table.head", "batch_id=abc123", "n_rows=5"],
        "metric_domain_kwargs": {"batch_id": "abc123"},
        "metric_domain_kwargs_id": "batch_id=abc123",
        "metric_name": "table.head",
        "metric_value_kwargs": {"n_rows": 5},
        "metric_value_kwargs_id": "n_rows=5",
    }


@pytest.mark.unit
def test_metric_configuration_get_domain_type(
    table_head_metric_config: MetricConfiguration,
    column_histogram_metric_config: MetricConfiguration,
) -> None:
    assert table_head_metric_config.get_domain_type() == MetricDomainTypes.TABLE
    assert column_histogram_metric_config.get_domain_type() == MetricDomainTypes.COLUMN


@pytest.mark.unit
def test_metric_configuration_get_domain(
    table_head_metric_config: MetricConfiguration,
    column_histogram_metric_config: MetricConfiguration,
) -> None:
    assert table_head_metric_config.get_domain() == Domain(
        domain_type=MetricDomainTypes.TABLE,
        domain_kwargs={},
    )
    assert column_histogram_metric_config.get_domain() == Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs={
            "column": "my_column",
        },
    )



from typing import Mapping, Sequence

import pandas as pd

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

MULTI_ASSET_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    DatabricksDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
]


@parameterize_batch_for_data_sources(
    data_source_configs=MULTI_ASSET_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2, 3, 4]}),
    extra_data={"other_table": pd.DataFrame({"col_b": ["a", "b", "c", "d"]})},
)
def test_success(
    batch_for_datasource: Batch,
    extra_table_names_for_datasource: Mapping[str, str],
):
    other_table_name = extra_table_names_for_datasource["other_table"]
    expectation = gxe.ExpectTableRowCountToEqualOtherTable(other_table_name=other_table_name)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=MULTI_ASSET_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2, 3, 4]}),
    extra_data={"other_table": pd.DataFrame({"col_b": ["just_this_one!"]})},
)
def test_different_counts(
    batch_for_datasource: Batch,
    extra_table_names_for_datasource: Mapping[str, str],
):
    other_table_name = extra_table_names_for_datasource["other_table"]
    expectation = gxe.ExpectTableRowCountToEqualOtherTable(other_table_name=other_table_name)
    result = batch_for_datasource.validate(expectation)
    assert not result.success
    assert result.result["observed_value"] == {
        "self": 4,
        "other": 1,
    }


@parameterize_batch_for_data_sources(
    data_source_configs=MULTI_ASSET_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2, 3, 4]}),
)
def test_missing_table(batch_for_datasource: Batch):
    expectation = gxe.ExpectTableRowCountToEqualOtherTable(other_table_name="where_am_i")
    result = batch_for_datasource.validate(expectation)
    assert not result.success, "We should not find the other table, since we didn't load it."


@parameterize_batch_for_data_sources(
    data_source_configs=MULTI_ASSET_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2, 3, 4]}),
    extra_data={"other_table": pd.DataFrame({"col_b": ["a", "b", "c", "d"]})},
)
def test_success_with_suite_param_other_table_name_(
    batch_for_datasource: Batch, extra_table_names_for_datasource: Mapping[str, str]
) -> None:
    suite_param_key = "test_expect_table_row_count_to_equal_other_table"
    other_table_name = extra_table_names_for_datasource["other_table"]
    expectation = gxe.ExpectTableRowCountToEqualOtherTable(
        other_table_name={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: other_table_name}
    )
    assert result.success



from typing import Sequence

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

COL_A = "col_a"
COL_B = "col_b"


DATA = pd.DataFrame(
    {
        COL_A: ["aa", "ab", "ac", None],
        COL_B: ["aa", "bb", "cc", None],
    }
)

SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    DatabricksDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


class TestNormalSql:
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(column=COL_A, like_pattern="z%"),
                id="no_matches",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(column=COL_A, like_pattern="_______"),
                id="too_many_underscores",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(
                    column=COL_B, like_pattern="a%", mostly=0.6
                ),
                id="mostly",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
    def test_success(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePattern,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert result.success

    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(column=COL_A, like_pattern="a%"),
                id="all_matches",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(column=COL_A, like_pattern="__"),
                id="underscores_match",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(
                    column=COL_B, like_pattern="a%", mostly=0.7
                ),
                id="mostly_threshold_not_met",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
    def test_failure(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePattern,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert not result.success


class TestMSSQL:
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(column=COL_A, like_pattern="a[xzy]"),
                id="bracket_notation",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=[MSSQLDatasourceTestConfig()], data=DATA
    )
    def test_success(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePattern,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert result.success

    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchLikePattern(column=COL_A, like_pattern="a[abc]"),
                id="bracket_notation_fail",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=[MSSQLDatasourceTestConfig()], data=DATA
    )
    def test_failure(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchLikePattern,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert not result.success



from typing import List

import pytest

from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.partitioners import ColumnPartitionerYearly
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.datasource.fluent.fluent_base_model import FluentBaseModel
from great_expectations.datasource.fluent.interfaces import Batch, DataAsset, Datasource
from great_expectations.datasource.fluent.pandas_datasource import PandasDatasource
from tests.conftest import random_name

DATASOURCE_NAME = "my datasource for batch configs"
EMPTY_DATA_ASSET_NAME = "my data asset for batch configs"
DATA_ASSET_WITH_BATCH_DEFINITION_NAME = "I have batch configs"
ANOTHER_DATA_ASSET_WITH_BATCH_DEFINITION_NAME = "I have batch configs too"
BATCH_DEFINITION_NAME = "my batch config"


@pytest.fixture
def file_context(empty_data_context: AbstractDataContext) -> AbstractDataContext:
    return empty_data_context


@pytest.fixture
def file_context_with_assets(file_context: AbstractDataContext) -> AbstractDataContext:
    """Context with a datasource that has 2 assets. one of the assets has a batch config."""
    datasource = file_context.data_sources.add_pandas(DATASOURCE_NAME)
    datasource.add_csv_asset(EMPTY_DATA_ASSET_NAME, "taxi.csv")  # type: ignore [arg-type]
    datasource.add_csv_asset(
        DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
        "taxi.csv",  # type: ignore [arg-type]
    ).add_batch_definition(BATCH_DEFINITION_NAME)
    datasource.add_csv_asset(
        ANOTHER_DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
        "taxi.csv",  # type: ignore [arg-type]
    ).add_batch_definition(BATCH_DEFINITION_NAME)

    return file_context


@pytest.fixture
def cloud_context(empty_cloud_context_fluent: CloudDataContext) -> AbstractDataContext:
    datasource = empty_cloud_context_fluent.data_sources.add_pandas(DATASOURCE_NAME)
    datasource.add_csv_asset(EMPTY_DATA_ASSET_NAME, "taxi.csv")  # type: ignore [arg-type]
    datasource.add_csv_asset(
        DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
        "taxi.csv",  # type: ignore [arg-type]
    ).add_batch_definition(BATCH_DEFINITION_NAME)
    datasource.add_csv_asset(
        ANOTHER_DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
        "taxi.csv",  # type: ignore [arg-type]
    ).add_batch_definition(BATCH_DEFINITION_NAME)
    return empty_cloud_context_fluent


@pytest.fixture
def datasource(file_context_with_assets: AbstractDataContext) -> PandasDatasource:
    output = file_context_with_assets.data_sources.get(DATASOURCE_NAME)
    assert isinstance(output, PandasDatasource)
    return output


@pytest.fixture
def empty_data_asset(datasource: Datasource) -> DataAsset:
    return datasource.get_asset(EMPTY_DATA_ASSET_NAME)


@pytest.fixture
def data_asset_with_batch_definition(datasource: Datasource) -> DataAsset:
    return datasource.get_asset(DATA_ASSET_WITH_BATCH_DEFINITION_NAME)


@pytest.fixture
def persisted_batch_definition(data_asset_with_batch_definition: DataAsset) -> BatchDefinition:
    return data_asset_with_batch_definition.batch_definitions[0]


@pytest.mark.unit
def test_add_batch_definition__success(empty_data_asset: DataAsset):
    name = "my batch config"
    batch_definition = empty_data_asset.add_batch_definition(name)

    assert batch_definition.name == name
    assert batch_definition.data_asset == empty_data_asset
    assert empty_data_asset.batch_definitions == [batch_definition]


@pytest.mark.unit
def test_add_batch_definition_with_partitioner__success(empty_data_asset: DataAsset):
    name = "my batch config"
    partitioner = ColumnPartitionerYearly(column_name="test-column")
    batch_definition = empty_data_asset.add_batch_definition(name, partitioner=partitioner)

    assert batch_definition.partitioner == partitioner


@pytest.mark.unit
def test_add_batch_definition__persists(
    file_context: AbstractDataContext, empty_data_asset: DataAsset
):
    name = "my batch config"
    partitioner = ColumnPartitionerYearly(column_name="test-column")
    batch_definition = empty_data_asset.add_batch_definition(name, partitioner=partitioner)

    loaded_datasource = file_context.data_sources.get(DATASOURCE_NAME)
    assert isinstance(loaded_datasource, Datasource)
    loaded_asset = loaded_datasource.get_asset(EMPTY_DATA_ASSET_NAME)

    assert loaded_asset.batch_definitions == [batch_definition]


@pytest.mark.unit
def test_add_batch_definition_with_partitioner__persists(
    file_context: AbstractDataContext, empty_data_asset: DataAsset
):
    name = "my batch config"
    partitioner = ColumnPartitionerYearly(column_name="test-column")
    empty_data_asset.add_batch_definition(name, partitioner=partitioner)

    loaded_datasource = file_context.data_sources.get(DATASOURCE_NAME)
    assert isinstance(loaded_datasource, Datasource)
    loaded_asset = loaded_datasource.get_asset(EMPTY_DATA_ASSET_NAME)

    loaded_batch_definition = loaded_asset.batch_definitions[0]
    assert loaded_batch_definition.partitioner == partitioner


@pytest.mark.unit
def test_add_batch_definition__multiple(empty_data_asset: DataAsset):
    empty_data_asset.add_batch_definition("foo")
    empty_data_asset.add_batch_definition("bar")

    assert len(empty_data_asset.batch_definitions) == 2


@pytest.mark.unit
def test_add_batch_definition__duplicate_key(empty_data_asset: DataAsset):
    name = "my batch config"
    empty_data_asset.add_batch_definition(name)

    with pytest.raises(ValueError, match="already exists"):
        empty_data_asset.add_batch_definition(name)


@pytest.mark.unit
def test_add_batch_definition__file_data__does_not_clobber_other_assets(
    file_context_with_assets: AbstractDataContext,
):
    _test_add_batch_definition__does_not_clobber_other_assets(
        context=file_context_with_assets,
        datasource_name=DATASOURCE_NAME,
    )


@pytest.mark.unit
def test_add_batch_definition__cloud_data__does_not_clobber_other_assets(
    cloud_context: AbstractDataContext,
):
    _test_add_batch_definition__does_not_clobber_other_assets(
        context=cloud_context,
        datasource_name=DATASOURCE_NAME,
    )


def _test_add_batch_definition__does_not_clobber_other_assets(
    context: AbstractDataContext,
    datasource_name: str,
):
    ds1 = context.data_sources.get(datasource_name)
    assert isinstance(ds1, PandasDatasource)
    my_asset = ds1.add_csv_asset("my asset", "taxi.csv")  # type: ignore [arg-type]

    # adding an asset currently clobbers the datasource, so for now we
    # need to reload the datasource AFTER adding the asset
    # TODO: Move fetching ds2 up with ds1 after V1-124
    ds2 = context.data_sources.get(datasource_name)
    assert isinstance(ds2, PandasDatasource)
    your_asset = ds2.add_csv_asset("your asset", "taxi.csv")  # type: ignore [arg-type]

    my_batch_definition = my_asset.add_batch_definition("my batch config")
    your_batch_definition = your_asset.add_batch_definition("your batch config")

    loaded_datasource = context.data_sources.get(datasource_name)
    assert isinstance(loaded_datasource, Datasource)
    assert loaded_datasource.get_asset(my_asset.name).batch_definitions == [my_batch_definition]
    assert loaded_datasource.get_asset(your_asset.name).batch_definitions == [your_batch_definition]


@pytest.mark.unit
def test_add_batch_definition__file_data__does_not_clobber_other_batch_definitions(
    file_context_with_assets: AbstractDataContext,
):
    _test_add_batch_definition__does_not_clobber_other_batch_definitions(
        context=file_context_with_assets,
        datasource_name=DATASOURCE_NAME,
        asset_name=EMPTY_DATA_ASSET_NAME,
    )


@pytest.mark.unit
def test_add_batch_definition__cloud_data__does_not_clobber_other_batch_definitions(
    cloud_context: AbstractDataContext,
):
    _test_add_batch_definition__does_not_clobber_other_batch_definitions(
        context=cloud_context,
        datasource_name=DATASOURCE_NAME,
        asset_name=EMPTY_DATA_ASSET_NAME,
    )


def _test_add_batch_definition__does_not_clobber_other_batch_definitions(
    context: AbstractDataContext,
    datasource_name: str,
    asset_name: str,
):
    ds1 = context.data_sources.get(datasource_name)
    ds2 = context.data_sources.get(datasource_name)
    assert isinstance(ds1, PandasDatasource)
    assert isinstance(ds2, PandasDatasource)

    asset_1 = ds1.get_asset(asset_name)
    asset_2 = ds2.get_asset(asset_name)

    my_batch_definition = asset_1.add_batch_definition("my batch config")
    your_batch_definition = asset_2.add_batch_definition("your batch config")

    loaded_datasource = context.data_sources.get(datasource_name)
    assert isinstance(loaded_datasource, Datasource)
    assert loaded_datasource.get_asset(asset_name).batch_definitions == [
        my_batch_definition,
        your_batch_definition,
    ]


@pytest.mark.unit
def test_delete_batch_definition__success(
    data_asset_with_batch_definition: DataAsset,
    persisted_batch_definition: BatchDefinition,
):
    assert persisted_batch_definition in data_asset_with_batch_definition.batch_definitions

    data_asset_with_batch_definition.delete_batch_definition(persisted_batch_definition.name)

    assert data_asset_with_batch_definition.batch_definitions == []


@pytest.mark.unit
def test_delete_batch_definition__persists(
    file_context_with_assets: AbstractDataContext,
    data_asset_with_batch_definition: DataAsset,
    persisted_batch_definition: BatchDefinition,
):
    data_asset_with_batch_definition.delete_batch_definition(persisted_batch_definition.name)

    loaded_datasource = file_context_with_assets.data_sources.get(DATASOURCE_NAME)
    assert isinstance(loaded_datasource, Datasource)
    loaded_asset = loaded_datasource.get_asset(EMPTY_DATA_ASSET_NAME)

    assert loaded_asset.batch_definitions == []


@pytest.mark.unit
def test_delete_batch_definition__unsaved_batch_definition(empty_data_asset: DataAsset):
    batch_definition = BatchDefinition[None](name="uh oh")

    with pytest.raises(ValueError, match="does not exist"):
        empty_data_asset.delete_batch_definition(batch_definition.name)


@pytest.mark.unit
def test_fields_set(empty_data_asset: DataAsset):
    """We mess with pydantic's internal __fields_set__ to determine
    if certain fields (batch_definitions in this case) should get serialized.

    This test is essentially a proxy for whether we serialize that field
    """
    asset = empty_data_asset

    # when we don't have batch configs, it shouldn't be in the set
    assert "batch_definitions" not in asset.__fields_set__

    # add some batch configs and ensure we have it in the set
    batch_definition_a = asset.add_batch_definition("a")
    batch_definition_b = asset.add_batch_definition("b")
    assert "batch_definitions" in asset.__fields_set__

    # delete one of the batch configs and ensure we still have it in the set
    asset.delete_batch_definition(batch_definition_a.name)
    assert "batch_definitions" in asset.__fields_set__

    # delete the remaining batch config and ensure we don't have it in the set
    asset.delete_batch_definition(batch_definition_b.name)
    assert "batch_definitions" not in asset.__fields_set__


@pytest.mark.unit
def test_delete_batch_definition__file_data__does_not_clobber_other_assets(
    file_context_with_assets: AbstractDataContext,
):
    _test_delete_batch_definition__does_not_clobber_other_assets(
        context=file_context_with_assets,
        datasource_name=DATASOURCE_NAME,
        asset_names=[
            DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
            ANOTHER_DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
        ],
    )


@pytest.mark.unit
def test_delete_batch_definition__cloud_data__does_not_clobber_other_assets(
    cloud_context: AbstractDataContext,
):
    _test_delete_batch_definition__does_not_clobber_other_assets(
        context=cloud_context,
        datasource_name=DATASOURCE_NAME,
        asset_names=[
            DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
            ANOTHER_DATA_ASSET_WITH_BATCH_DEFINITION_NAME,
        ],
    )


def _test_delete_batch_definition__does_not_clobber_other_assets(
    context: AbstractDataContext, datasource_name: str, asset_names: List[str]
):
    # each asset has one batch config; delete it
    for asset_name in asset_names:
        datasource = context.data_sources.get(datasource_name)
        assert isinstance(datasource, Datasource)
        asset = datasource.get_asset(asset_name)
        asset.delete_batch_definition(asset.batch_definitions[0].name)

    loaded_datasource = context.data_sources.get(datasource_name)
    assert isinstance(loaded_datasource, Datasource)

    # ensure neither call to delete_batch_definition() didn't clobber each other
    for asset_name in asset_names:
        assert loaded_datasource.get_asset(asset_name).batch_definitions == []


class _MyPartitioner(FluentBaseModel):
    """Partitioner that adhere's to the expected protocol."""

    sort_ascending: bool = True

    @property
    def param_names(self) -> List[str]:
        return ["a", "b"]


@pytest.fixture
def metadata_1_1(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": 1, "b": 1})


@pytest.fixture
def metadata_1_2(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": 1, "b": 2})


@pytest.fixture
def metadata_2_1(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": 2, "b": 1})


@pytest.fixture
def metadata_2_2(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": 2, "b": 2})


@pytest.fixture
def metadata_2_none(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": 2, "b": None})


@pytest.fixture
def metadata_none_2(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": None, "b": 2})


@pytest.fixture
def metadata_none_none(mocker):
    return mocker.MagicMock(spec=Batch, metadata={"a": None, "b": None})


@pytest.mark.unit
def test_sort_batches__ascending(
    empty_data_asset,
    metadata_1_1,
    metadata_1_2,
    metadata_2_1,
    metadata_2_2,
    metadata_none_2,
    metadata_2_none,
    metadata_none_none,
):
    partitioner = _MyPartitioner(sort_ascending=True)
    batches = [
        metadata_1_1,
        metadata_1_2,
        metadata_2_1,
        metadata_2_2,
        metadata_2_none,
        metadata_none_2,
        metadata_none_none,
    ]

    batches = empty_data_asset.sort_batches(batches, partitioner)

    assert batches == [
        metadata_none_none,
        metadata_none_2,
        metadata_1_1,
        metadata_1_2,
        metadata_2_none,
        metadata_2_1,
        metadata_2_2,
    ]


@pytest.mark.unit
def test_sort_batches__descending(
    empty_data_asset,
    metadata_1_1,
    metadata_1_2,
    metadata_2_1,
    metadata_2_2,
    metadata_none_2,
    metadata_2_none,
    metadata_none_none,
):
    partitioner = _MyPartitioner(sort_ascending=False)
    batches = [
        metadata_1_1,
        metadata_1_2,
        metadata_2_1,
        metadata_2_2,
        metadata_2_none,
        metadata_none_2,
        metadata_none_none,
    ]

    batches = empty_data_asset.sort_batches(batches, partitioner)

    assert batches == [
        metadata_2_2,
        metadata_2_1,
        metadata_2_none,
        metadata_1_2,
        metadata_1_1,
        metadata_none_2,
        metadata_none_none,
    ]


@pytest.mark.unit
def test_sort_batches__requires_keys(empty_data_asset, mocker):
    partitioner = _MyPartitioner()

    wheres_my_b = mocker.MagicMock(spec=Batch, metadata={"a": 1})
    i_have_a_b = mocker.MagicMock(spec=Batch, metadata={"a": 1, "b": 2})

    expected_error = "Trying to sort my data asset for batch configs's batches on key b"

    with pytest.raises(KeyError, match=expected_error):
        empty_data_asset.sort_batches([wheres_my_b, i_have_a_b], partitioner)


class TestGetBatchDefinition:
    def test_get_returns_id(self, data_context: AbstractDataContext) -> None:
        # arrange
        resource_name = random_name()
        ds = data_context.data_sources.add_pandas(
            name=resource_name,
        )
        asset = ds.add_dataframe_asset(name=resource_name)
        batch_def = asset.add_batch_definition_whole_dataframe(name=resource_name)

        # act
        refetched_batch_def = asset.get_batch_definition(resource_name)

        # assert
        assert batch_def.id
        assert refetched_batch_def.id
        assert batch_def.id == refetched_batch_def.id



from typing import List

from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

postgresql_integration_tests = []


connecting_to_your_data = [
    IntegrationTestFixture(
        name="how_to_configure_credentials",
        user_flow_script="docs/docusaurus/docs/oss/guides/setup/configuring_data_contexts/how_to_configure_credentials.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
        backend_dependencies=[BackendDependencies.POSTGRESQL],
    ),
]

partition_data = [
    IntegrationTestFixture(
        name="partition_data_on_whole_table_postgres",
        user_flow_script="tests/integration/db/test_sql_data_partitioned_on_whole_table.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        util_script="tests/test_utils.py",
        other_files=(
            (
                "tests/integration/fixtures/partition_and_sample_data/postgres_connection_string.yml",
                "connection_string.yml",
            ),
        ),
        backend_dependencies=[BackendDependencies.POSTGRESQL],
    ),
    IntegrationTestFixture(
        name="partition_data_on_datetime_postgres",
        user_flow_script="tests/integration/db/test_sql_data_partitioned_on_datetime_and_day_part.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        util_script="tests/test_utils.py",
        other_files=(
            (
                "tests/integration/fixtures/partition_and_sample_data/postgres_connection_string.yml",
                "connection_string.yml",
            ),
        ),
        backend_dependencies=[BackendDependencies.POSTGRESQL],
    ),
]

sample_data: List[IntegrationTestFixture] = []


fluent_datasources = [
    IntegrationTestFixture(
        name="how_to_connect_to_postgresql_data",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/database/how_to_connect_to_postgresql_data.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.POSTGRESQL],
    ),
]

getting_started = []

unexpected_row_expectation = [
    IntegrationTestFixture(
        name="how_to_write_an_unexpected_row_expectation",
        user_flow_script="docs/docusaurus/docs/snippets/unexpected_row_expectation.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
        util_script="tests/test_utils.py",
        backend_dependencies=[BackendDependencies.POSTGRESQL],
    )
]

postgresql_integration_tests += connecting_to_your_data
postgresql_integration_tests += partition_data
postgresql_integration_tests += sample_data
postgresql_integration_tests += fluent_datasources
postgresql_integration_tests += getting_started
postgresql_integration_tests += unexpected_row_expectation



import pandas as pd
import pytest

import great_expectations as gx
from great_expectations.compatibility.pydantic import ValidationError
from great_expectations.core.types import Comparable
from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.column.mean import ColumnMean
from great_expectations.metrics.metric import AbstractClassInstantiationError, Metric
from great_expectations.metrics.metric_results import MetricResult
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,
    MetricConfigurationID,
)

BATCH_ID = "my_data_source-my_data_asset-year_2025"
COLUMN = "my_column"

FULLY_QUALIFIED_METRIC_NAME = "column_values.above"


class ColumnValuesAboveResult(MetricResult[bool]): ...


class ColumnValuesAbove(ColumnMetric[ColumnValuesAboveResult]):
    name = FULLY_QUALIFIED_METRIC_NAME

    min_value: Comparable
    strict_min: bool = False


class TestMetric:
    @pytest.mark.unit
    def test_metric_instantiation_raises(self):
        with pytest.raises(AbstractClassInstantiationError):
            Metric(column=COLUMN)


class TestMetricDefinition:
    @pytest.mark.unit
    def test_success(self):
        class MyColumnValuesAbove(ColumnMetric[ColumnValuesAboveResult]):
            name = FULLY_QUALIFIED_METRIC_NAME

            min_value: Comparable
            strict_min: bool = False

    @pytest.mark.unit
    def test_success_without_generic_return_types(self):
        class MyColumnValuesAbove(ColumnMetric):
            name = FULLY_QUALIFIED_METRIC_NAME

            min_value: Comparable
            strict_min: bool = False


class TestMetricInstantiation:
    @pytest.mark.unit
    def test_instantiation_success(self):
        ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )

    @pytest.mark.unit
    def test_instantiation_missing_domain_parameters_raises(self):
        with pytest.raises(ValidationError):
            ColumnValuesAbove(min_value=42)


class TestMetricConfig:
    @pytest.mark.unit
    def test_success(self):
        expected_config = MetricConfiguration(
            metric_name=FULLY_QUALIFIED_METRIC_NAME,
            metric_domain_kwargs={
                "batch_id": BATCH_ID,
                "row_condition": None,
                "condition_parser": None,
                "column": COLUMN,
            },
            metric_value_kwargs={
                "min_value": 42,
                "strict_min": False,
            },
        )

        actual_config = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        ).config(batch_id=BATCH_ID)

        assert actual_config.metric_name == expected_config.metric_name
        assert actual_config.metric_domain_kwargs == expected_config.metric_domain_kwargs
        assert actual_config.metric_value_kwargs == expected_config.metric_value_kwargs
        assert isinstance(actual_config.id, MetricConfigurationID)


class TestMetricImmutability:
    @pytest.mark.unit
    def test_domain_kwarg_immutability_success(self):
        column_values_above = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )

        with pytest.raises(TypeError):
            column_values_above.column = "updated_column"

    @pytest.mark.unit
    def test_value_kwarg_immutability_success(self):
        column_values_above = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )

        with pytest.raises(TypeError):
            column_values_above.min_value = 42


class TestMetricIdFromBatch:
    @pytest.mark.unit
    def test_two_identical_metrics_with_same_batch_id_returns_same_metric_id(self):
        metric_1 = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )
        metric_2 = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )
        assert metric_1.metric_id_for_batch(BATCH_ID) == metric_2.metric_id_for_batch(BATCH_ID)

    @pytest.mark.unit
    def test_metric_with_different_value_kwargs_return_different_ids(self):
        metric_1 = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )
        metric_2 = ColumnValuesAbove(
            column=COLUMN,
            min_value=43,
        )
        assert metric_1.metric_id_for_batch(BATCH_ID) != metric_2.metric_id_for_batch(BATCH_ID)

    @pytest.mark.unit
    def test_metric_with_different_batch_ids_return_different_ids(self):
        metric_1 = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )
        metric_2 = ColumnValuesAbove(
            column=COLUMN,
            min_value=42,
        )
        assert metric_1.metric_id_for_batch(BATCH_ID) != metric_2.metric_id_for_batch(
            BATCH_ID + "_another"
        )


class TestComputeMetric:
    @pytest.mark.unit
    def test_same_metric_different_args_have_different_results(self):
        context = gx.get_context(mode="ephemeral")
        data_source = context.data_sources.add_pandas("Pandas Data Source")
        data_asset = data_source.add_dataframe_asset("DataFrame Asset")
        batch_definition = data_asset.add_batch_definition_whole_dataframe(
            "Whole DataFrame Batch Definition"
        )
        df = pd.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4, 5, 6],
            }
        )
        batch = batch_definition.get_batch({"dataframe": df})
        metrics = [
            ColumnMean(column="a"),
            ColumnMean(column="b"),
        ]
        metric_result = batch.compute_metrics(metrics)
        assert isinstance(metric_result, list)
        assert len(metric_result) == 2
        assert metric_result[0].value == 2.0
        assert metric_result[0].id == metrics[0].metric_id_for_batch(batch.id)
        assert metric_result[1].value == 5.0
        assert metric_result[1].id == metrics[1].metric_id_for_batch(batch.id)
        assert metric_result[0].id != metric_result[1].id

    @pytest.mark.unit
    def test_single_metric_as_list_result_is_list(self):
        context = gx.get_context(mode="ephemeral")
        data_source = context.data_sources.add_pandas("Pandas Data Source")
        data_asset = data_source.add_dataframe_asset("DataFrame Asset")
        batch_definition = data_asset.add_batch_definition_whole_dataframe(
            "Whole DataFrame Batch Definition"
        )
        df = pd.DataFrame(
            {
                "a": [1, 2, 3],
            }
        )
        batch = batch_definition.get_batch({"dataframe": df})
        metrics = [
            ColumnMean(column="a"),
        ]
        metric_result = batch.compute_metrics(metrics)
        assert isinstance(metric_result, list)
        assert len(metric_result) == 1
        assert metric_result[0].value == 2.0

    @pytest.mark.unit
    def test_single_metric_result_is_metric_result(self):
        context = gx.get_context(mode="ephemeral")
        data_source = context.data_sources.add_pandas("Pandas Data Source")
        data_asset = data_source.add_dataframe_asset("DataFrame Asset")
        batch_definition = data_asset.add_batch_definition_whole_dataframe(
            "Whole DataFrame Batch Definition"
        )
        df = pd.DataFrame(
            {
                "a": [1, 2, 3],
            }
        )
        batch = batch_definition.get_batch({"dataframe": df})
        metric_result = batch.compute_metrics(ColumnMean(column="a"))
        assert isinstance(metric_result, MetricResult)
        assert metric_result.value == 2.0



from __future__ import annotations

from typing import Any, Dict

from great_expectations.compatibility import pyspark
from great_expectations.compatibility.pyspark import functions as F
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import MetricPartialFunctionTypes
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnMapMetricProvider,
    column_condition_partial,
)
from great_expectations.expectations.metrics.metric_provider import metric_partial


class ColumnValuesDecreasing(ColumnMapMetricProvider):
    condition_metric_name = "column_values.decreasing"
    condition_value_keys = ("strictly",)
    default_kwarg_values = {
        "strictly": False,
    }

    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        column,
        **kwargs,
    ):
        temp_column = column

        series_diff = temp_column.diff()
        series_diff[series_diff.isnull()] = -1

        strictly: bool = kwargs.get("strictly") or False
        if strictly:
            return series_diff < 0
        return series_diff <= 0

    @metric_partial(
        engine=SparkDFExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.WINDOW_CONDITION_FN,
        domain_type=MetricDomainTypes.COLUMN,
    )
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        # check if column is any type that could have na (numeric types)
        column_name = metric_domain_kwargs["column"]
        table_columns = metrics["table.column_types"]
        column_metadata = [col for col in table_columns if col["name"] == column_name][0]
        if isinstance(
            column_metadata["type"],
            (
                pyspark.types.LongType,
                pyspark.types.DoubleType,
                pyspark.types.IntegerType,
            ),
        ):
            # if column is any type that could have NA values, remove them (not filtered by .isNotNull())  # noqa: E501 # FIXME CoP
            compute_domain_kwargs = execution_engine.add_column_row_condition(
                metric_domain_kwargs,
                filter_null=cls.filter_column_isnull,
                filter_nan=True,
            )
        else:
            compute_domain_kwargs = metric_domain_kwargs

        (
            _df,
            compute_domain_kwargs,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(compute_domain_kwargs, MetricDomainTypes.COLUMN)

        # instead detect types naturally
        column = F.col(column_name)
        if isinstance(
            column_metadata["type"],
            (pyspark.types.TimestampType, pyspark.types.DateType),
        ):
            diff = F.datediff(
                column,
                F.lag(column).over(pyspark.Window.orderBy(F.lit("constant"))),
            )
        else:
            diff = column - F.lag(column).over(pyspark.Window.orderBy(F.lit("constant")))
            diff = F.when(diff.isNull(), -1).otherwise(diff)

        # NOTE: because in spark we are implementing the window function directly,
        # we have to return the *unexpected* condition
        if metric_value_kwargs["strictly"]:
            return (
                F.when(diff >= 0, F.lit(True)).otherwise(F.lit(False)),
                compute_domain_kwargs,
                accessor_domain_kwargs,
            )
        # If we expect values to be flat or decreasing then unexpected values are those
        # that are decreasing
        else:
            return (
                F.when(diff > 0, F.lit(True)).otherwise(F.lit(False)),
                compute_domain_kwargs,
                accessor_domain_kwargs,
            )



import pandas as pd
import pytest

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column_pair.values_in_set import (
    ColumnPairValuesInSetUnexpectedCount,
    ColumnPairValuesInSetUnexpectedCountResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import SparkFilesystemCsvDatasourceTestConfig
from tests.metrics.conftest import PANDAS_DATA_SOURCES, SPARK_DATA_SOURCES, SQL_DATA_SOURCES

COL_A = "A"
COL_B = "B"
COL_A_WITH_NULLS = "A_WITH_NULLS"
COL_B_WITH_NULLS = "B_WITH_NULLS"
DATA_FRAME = pd.DataFrame(
    {
        "id": [1, 2, 3, 4],
        COL_A: [1, 2, 3, 4],
        COL_B: ["a", "b", "c", "d"],
    },
)
DATA_FRAME_WITH_NULLS = pd.DataFrame(
    {
        COL_A_WITH_NULLS: [None, 2, None, 4, None, None],
        COL_B_WITH_NULLS: ["a", "b", None, None, None, None],
    }
)
SUCCESS_VALUE_PAIR_SET = {(1, "a"), (2, "b"), (3, "c"), (4, "d")}
FAILURE_VALUE_PAIR_SET = {(1, "a"), (2, "b"), (3, "c")}
FAILURE_RESULT_COUNT = 1
NO_MATCH_PAIR_SET = {(5, "e")}


class TestColumnPairValuesInSetUnexpectedValues:
    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES + SPARK_DATA_SOURCES + SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnPairValuesInSetUnexpectedCount(
            value_pairs_set=SUCCESS_VALUE_PAIR_SET,
            column_A=COL_A,
            column_B=COL_B,
        )
        result = batch.compute_metrics(metric)
        assert isinstance(result, ColumnPairValuesInSetUnexpectedCountResult)
        assert result.value == 0

    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES + SPARK_DATA_SOURCES + SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_failure(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnPairValuesInSetUnexpectedCount(
            value_pairs_set=FAILURE_VALUE_PAIR_SET,
            column_A=COL_A,
            column_B=COL_B,
        )
        result = batch.compute_metrics(metric)
        assert isinstance(result, ColumnPairValuesInSetUnexpectedCountResult)
        assert result.value == FAILURE_RESULT_COUNT

    @pytest.mark.parametrize(
        "ignore_row_if,unexpected_count",
        [
            ("either_value_is_missing", 1),
            ("both_values_are_missing", 3),
            ("neither", 6),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=DATA_FRAME_WITH_NULLS,
    )
    def test_ignore_row_if__pandas(
        self, batch_for_datasource: Batch, ignore_row_if, unexpected_count
    ) -> None:
        batch = batch_for_datasource
        metric = ColumnPairValuesInSetUnexpectedCount(
            value_pairs_set=NO_MATCH_PAIR_SET,
            column_A=COL_A_WITH_NULLS,
            column_B=COL_B_WITH_NULLS,
            ignore_row_if=ignore_row_if,
        )
        result = batch.compute_metrics(metric)
        assert result.value == unexpected_count

    @pytest.mark.parametrize(
        "ignore_row_if,unexpected_count",
        [
            pytest.param(
                "either_value_is_missing",
                1,
                marks=pytest.mark.xfail(reason="returns 2", strict=True),
            ),
            pytest.param(
                "both_values_are_missing",
                3,
                marks=pytest.mark.xfail(reason="returns 6", strict=True),
            ),
            pytest.param("neither", 6),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=[SparkFilesystemCsvDatasourceTestConfig()],
        data=DATA_FRAME_WITH_NULLS,
    )
    def test_ignore_row_if__spark(
        self, batch_for_datasource: Batch, ignore_row_if, unexpected_count
    ) -> None:
        """This test captures unexpected behavior for Spark FileSystem data sources."""
        batch = batch_for_datasource
        metric = ColumnPairValuesInSetUnexpectedCount(
            value_pairs_set=NO_MATCH_PAIR_SET,
            column_A=COL_A_WITH_NULLS,
            column_B=COL_B_WITH_NULLS,
            ignore_row_if=ignore_row_if,
        )
        result = batch.compute_metrics(metric)
        assert result.value == unexpected_count

    @pytest.mark.parametrize(
        "ignore_row_if,unexpected_count",
        [
            pytest.param("either_value_is_missing", 1),
            pytest.param("both_values_are_missing", 3),
            pytest.param(
                "neither",
                6,
                marks=pytest.mark.xfail(
                    reason=("returns 3 - pairs where both are null are dropped"),
                    strict=True,
                ),
            ),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME_WITH_NULLS,
    )
    def test_ignore_row_if__sql(
        self, batch_for_datasource: Batch, ignore_row_if, unexpected_count
    ) -> None:
        """This test captures a bug with SQL data sources and the ignore_row_if condition,
        where column pairs are dropped if both values are null.
        """
        batch = batch_for_datasource
        metric = ColumnPairValuesInSetUnexpectedCount(
            value_pairs_set=NO_MATCH_PAIR_SET,
            column_A=COL_A_WITH_NULLS,
            column_B=COL_B_WITH_NULLS,
            ignore_row_if=ignore_row_if,
        )
        result = batch.compute_metrics(metric)
        assert result.value == unexpected_count



from __future__ import annotations

import logging
import os
import pathlib
from typing import TYPE_CHECKING, Callable, ClassVar, List, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_connector import (
    FilePathDataConnector,
)

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr

logger = logging.getLogger(__name__)


class FilesystemOptions(pydantic.BaseModel):
    glob_directive: str = "**/*"


class FilesystemDataConnector(FilePathDataConnector):
    """Extension of FilePathDataConnector used to connect to Filesystem (local, networked file storage (NFS), DBFS, etc.).

    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
        base_directory: Relative path to subdirectory containing files of interest
        glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
        data_context_root_directory: Optional GreatExpectations root directory (if installed on filesystem)
        whole_directory_path_override: Treat an entire directory as a single Asset
    """  # noqa: E501 # FIXME CoP

    asset_level_option_keys: ClassVar[tuple[str, ...]] = ("glob_directive",)
    asset_options_type: ClassVar[Type[FilesystemOptions]] = FilesystemOptions

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_asset_name: str,
        base_directory: pathlib.Path,
        glob_directive: str = "**/*",
        data_context_root_directory: Optional[pathlib.Path] = None,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> None:
        self._base_directory = base_directory
        self._glob_directive: str = glob_directive
        self._data_context_root_directory: Optional[pathlib.Path] = data_context_root_directory

        super().__init__(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @property
    def base_directory(self) -> pathlib.Path:
        """
        Accessor method for base_directory. If directory is a relative path, interpret it as relative to the
        root directory. If it is absolute, then keep as-is.
        """  # noqa: E501 # FIXME CoP
        return normalize_directory_path(
            dir_path=self._base_directory,
            root_directory_path=self._data_context_root_directory,
        )

    @classmethod
    def build_data_connector(  # noqa: PLR0913 # FIXME CoP
        cls,
        datasource_name: str,
        data_asset_name: str,
        base_directory: pathlib.Path,
        glob_directive: str = "**/*",
        data_context_root_directory: Optional[pathlib.Path] = None,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> FilesystemDataConnector:
        """Builds "FilesystemDataConnector", which links named DataAsset to filesystem.

        Args:
            datasource_name: The name of the Datasource associated with this "FilesystemDataConnector" instance
            data_asset_name: The name of the DataAsset using this "FilesystemDataConnector" instance
            base_directory: Relative path to subdirectory containing files of interest
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            data_context_root_directory: Optional GreatExpectations root directory (if installed on filesystem)
            file_path_template_map_fn: Format function mapping path to fully-qualified resource on filesystem (optional)
            get_unfiltered_batch_definition_list_fn: Function used to get the batch definition list before filtering

        Returns:
            Instantiated "FilesystemDataConnector" object
        """  # noqa: E501 # FIXME CoP
        return FilesystemDataConnector(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            base_directory=base_directory,
            glob_directive=glob_directive,
            data_context_root_directory=data_context_root_directory,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_test_connection_error_message(
        cls,
        data_asset_name: str,
        base_directory: pathlib.Path,
        glob_directive: str = "**/*",
        data_context_root_directory: Optional[pathlib.Path] = None,
    ) -> str:
        """Builds helpful error message for reporting issues when linking named DataAsset to filesystem.

        Args:
            data_asset_name: The name of the DataAsset using this "FilesystemDataConnector" instance
            base_directory: Relative path to subdirectory containing files of interest
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            data_context_root_directory: Optional GreatExpectations root directory (if installed on filesystem)

        Returns:
            Customized error message
        """  # noqa: E501 # FIXME CoP
        test_connection_error_message_template: str = 'No file at base_directory path "{base_directory}" matched glob_directive "{glob_directive}" for DataAsset "{data_asset_name}".'  # noqa: E501 # FIXME CoP
        return test_connection_error_message_template.format(
            **{
                "data_asset_name": data_asset_name,
                "base_directory": base_directory.resolve(),
                "glob_directive": glob_directive,
                "data_context_root_directory": data_context_root_directory,
            }
        )

    # Interface Method
    @override
    def get_data_references(self) -> List[str]:
        base_directory: pathlib.Path = self.base_directory
        glob_directive: str = self._glob_directive
        path_list: List[str] = get_filesystem_one_level_directory_glob_path_list(
            base_directory_path=base_directory, glob_directive=glob_directive
        )
        return sorted(path_list)

    # Interface Method
    @override
    def _get_full_file_path(self, path: str) -> str:
        return str(self.base_directory.joinpath(path))


def normalize_directory_path(
    dir_path: Union[PathStr],
    root_directory_path: Optional[PathStr] = None,
) -> pathlib.Path:
    dir_path = pathlib.Path(dir_path)

    # If directory is a relative path, interpret it as relative to the root directory.
    if dir_path.is_absolute() or root_directory_path is None:
        return dir_path

    root_directory_path = pathlib.Path(root_directory_path)

    return root_directory_path.joinpath(dir_path)


def get_filesystem_one_level_directory_glob_path_list(
    base_directory_path: Union[PathStr], glob_directive: str
) -> List[str]:
    """
    List file names, relative to base_directory_path one level deep, with expansion specified by glob_directive.
    :param base_directory_path -- base directory path, relative to which file paths will be collected
    :param glob_directive -- glob expansion directive
    :returns -- list of relative file paths
    """  # noqa: E501 # FIXME CoP
    if isinstance(base_directory_path, str):
        base_directory_path = pathlib.Path(base_directory_path)

    globbed_paths = base_directory_path.glob(glob_directive)

    path_list: List[str] = [
        os.path.relpath(str(posix_path), base_directory_path) for posix_path in globbed_paths
    ]

    return path_list



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    Union,
    cast,
)

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.partitioners import PartitionerConvertedDatetime
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.sql_datasource import (
    QueryAsset as SqlQueryAsset,
)
from great_expectations.datasource.fluent.sql_datasource import (
    SQLDatasource,
    SqlitePartitionerConvertedDateTime,
    _PartitionerOneColumnOneParam,
)
from great_expectations.datasource.fluent.sql_datasource import (
    TableAsset as SqlTableAsset,
)
from great_expectations.execution_engine.sqlite_execution_engine import SqliteExecutionEngine

if TYPE_CHECKING:
    # min version of typing_extension missing `Self`, so it can't be imported at runtime

    from great_expectations.core.partitioners import (
        ColumnPartitioner,
    )
    from great_expectations.datasource.fluent.interfaces import (
        BatchMetadata,
        BatchParameters,
        DataAsset,
    )
    from great_expectations.execution_engine.sqlalchemy_execution_engine import (
        SqlAlchemyExecutionEngine,
    )

# This module serves as an example of how to extend _SQLAssets for specific backends. The steps are:
# 1. Create a plain class with the extensions necessary for the specific backend.
# 2. Make 2 classes XTableAsset and XQueryAsset by mixing in the class created in step 1 with
#    sql_datasource.TableAsset and sql_datasource.QueryAsset.
#
# See SqliteDatasource, SqliteTableAsset, and SqliteQueryAsset below.


class PartitionerConvertedDateTime(_PartitionerOneColumnOneParam):
    """A partitioner than can be used for sql engines that represents datetimes as strings.

    The SQL engine that this currently supports is SQLite since it stores its datetimes as
    strings.
    The DatetimePartitioner will also work for SQLite and may be more intuitive.
    """

    # date_format_strings syntax is documented here:
    # https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    # It allows for arbitrary strings so can't be validated until conversion time.
    date_format_string: str
    column_name: str
    method_name: Literal["partition_on_converted_datetime"] = "partition_on_converted_datetime"

    @property
    @override
    def param_names(self) -> List[str]:
        # The datetime parameter will be a string representing a datetime in the format
        # given by self.date_format_string.
        return ["datetime"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {
            "column_name": self.column_name,
            "date_format_string": self.date_format_string,
        }

    @override
    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        if "datetime" not in options:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "'datetime' must be specified in the batch parameters to create a batch identifier"
            )
        return {self.column_name: options["datetime"]}


class SqliteDsn(pydantic.AnyUrl):
    allowed_schemes = {
        "sqlite",
        "sqlite+pysqlite",
        "sqlite+aiosqlite",
        "sqlite+pysqlcipher",
    }
    host_required = False


class SqliteTableAsset(SqlTableAsset):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # update the partitioner map with the Sqlite specific partitioner
        self._partitioner_implementation_map[PartitionerConvertedDatetime] = (
            SqlitePartitionerConvertedDateTime
        )

    type: Literal["table"] = "table"

    @override
    def validate_batch_definition(self, partitioner: ColumnPartitioner) -> None:
        # TODO: Implement batch definition validation.
        # sqlite stores datetimes as a string so we must override how we normally
        # validate batch definitions.
        pass


class SqliteQueryAsset(SqlQueryAsset):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # update the partitioner map with the  Sqlite specific partitioner
        self._partitioner_implementation_map[PartitionerConvertedDatetime] = (
            SqlitePartitionerConvertedDateTime
        )

    type: Literal["query"] = "query"

    @override
    def validate_batch_definition(self, partitioner: ColumnPartitioner) -> None:
        # TODO: Implement batch definition validation.
        # sqlite stores datetimes as a string so we must override how we normally
        # validate batch definitions.
        pass


@public_api
class SqliteDatasource(SQLDatasource):
    """Adds a sqlite datasource to the data context.

    Args:
        name: The name of this sqlite datasource.
        connection_string: The SQLAlchemy connection string used to connect to the sqlite database.
            For example: "sqlite:///path/to/file.db"
        create_temp_table: Whether to leverage temporary tables during metric computation.
        assets: An optional dictionary whose keys are TableAsset names and whose values
            are TableAsset objects.
    """

    # class var definitions
    asset_types: ClassVar[List[Type[DataAsset]]] = [SqliteTableAsset, SqliteQueryAsset]

    # Subclass instance var overrides
    # right side of the operator determines the type name
    # left side enforces the names on instance creation
    type: Literal["sqlite"] = "sqlite"  # type: ignore[assignment] # FIXME CoP
    connection_string: Union[ConfigStr, SqliteDsn]

    _TableAsset: Type[SqlTableAsset] = pydantic.PrivateAttr(SqliteTableAsset)
    _QueryAsset: Type[SqlQueryAsset] = pydantic.PrivateAttr(SqliteQueryAsset)

    @property
    @override
    def execution_engine_type(self) -> Type[SqlAlchemyExecutionEngine]:
        """Returns the default execution engine type."""
        return SqliteExecutionEngine

    @public_api
    @override
    def add_table_asset(
        self,
        name: str,
        table_name: str = "",
        schema_name: Optional[str] = None,
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> SqliteTableAsset:
        """Adds a table asset to this SQLite datasource

        Args:
            name: The name of this table asset
            table_name: The name of the database table
            schema_name: The schema to which this table belongs
            batch_metadata: An arbitrary dictionary for a caller to annotate the asset

        Returns:
            The SqliteTableAsset added
        """
        return cast(
            "SqliteTableAsset",
            super().add_table_asset(
                name=name,
                table_name=table_name,
                schema_name=schema_name,
                batch_metadata=batch_metadata,
            ),
        )

    add_table_asset.__doc__ = SQLDatasource.add_table_asset.__doc__

    @public_api
    @override
    def add_query_asset(
        self,
        name: str,
        query: str,
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> SqliteQueryAsset:
        """Adds a query asset to this SQLite datasource

        Args:
            name: The name of this query asset
            query: The SQL query
            batch_metadata: An arbitrary dictionary for a caller to annotate the asset

        Returns:
            The SqliteQueryAsset added
        """

        return cast(
            "SqliteQueryAsset",
            super().add_query_asset(name=name, query=query, batch_metadata=batch_metadata),
        )

    add_query_asset.__doc__ = SQLDatasource.add_query_asset.__doc__



from typing import TYPE_CHECKING

import pytest
from contrib.experimental.great_expectations_experimental.expectations.expect_queried_column_value_frequency_to_meet_threshold import (  # noqa: E501 # FIXME CoP
    ExpectQueriedColumnValueFrequencyToMeetThreshold,  # noqa: F401 # needed for expectation registration
)

# noinspection PyUnresolvedReferences
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest
from great_expectations.self_check.util import (
    get_test_validator_with_data,
)
from great_expectations.util import build_in_memory_runtime_context
from great_expectations.validator.validator import (
    ExpectationValidationResult,
    Validator,
)

if TYPE_CHECKING:
    import pandas as pd

sqlite_runtime_batch_request: RuntimeBatchRequest = RuntimeBatchRequest(
    datasource_name="my_sqlite_db_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="titanic",
    runtime_parameters={"query": "SELECT * FROM titanic LIMIT 100"},
    batch_identifiers={"default_identifier_name": "test_identifier"},
    batch_spec_passthrough={"create_temp_table": False},
)

sqlite_batch_request: BatchRequest = BatchRequest(
    datasource_name="my_sqlite_db_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="titanic",
    batch_spec_passthrough={"create_temp_table": False},
)

pytest.skip("TODO: Fix in V1-323", allow_module_level=True)


@pytest.mark.parametrize(
    "batch_request,success,observed,row_condition,warns",
    [
        (sqlite_runtime_batch_request, True, 0.54, None, False),
        (sqlite_batch_request, True, 0.6481340441736482, None, False),
        (sqlite_batch_request, False, 0.4791666666666667, 'col("Age")<18', True),
        (sqlite_runtime_batch_request, True, 0.5, 'col("Age")>17', True),
    ],
)
@pytest.mark.slow  # 3.02s
@pytest.mark.filesystem
def test_expect_queried_column_value_frequency_to_meet_threshold_sqlite(
    batch_request,
    success,
    observed,
    row_condition,
    warns,
    titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    validator: Validator = context.get_validator(batch_request=batch_request)

    if warns:
        with pytest.warns(UserWarning):
            result: ExpectationValidationResult = (
                validator.expect_queried_column_value_frequency_to_meet_threshold(
                    column="Sex",
                    value="male",
                    threshold=0.5,
                    row_condition=row_condition,
                    condition_parser="great_expectations",
                )
            )
    else:
        result: ExpectationValidationResult = (
            validator.expect_queried_column_value_frequency_to_meet_threshold(
                column="Sex",
                value="male",
                threshold=0.5,
                row_condition=row_condition,
                condition_parser="great_expectations",
            )
        )
    assert result["success"] == success and result["result"]["observed_value"] == observed


@pytest.mark.parametrize(
    "batch_request,query,success,observed,row_condition,warns",
    [
        (
            sqlite_batch_request,
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT({col}) FROM titanic) FROM titanic GROUP BY {col}",  # noqa: E501 # FIXME CoP
            True,
            0.6481340441736482,
            None,
            True,
        ),
        (
            sqlite_runtime_batch_request,
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT({col}) FROM titanic) FROM {batch} GROUP BY {col}",  # noqa: E501 # FIXME CoP
            False,
            0.04112718964204113,
            None,
            True,
        ),
        (
            sqlite_batch_request,
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT(y) FROM wrong) FROM {batch} GROUP BY {col}",  # noqa: E501 # FIXME CoP
            True,
            7.091666666666667,
            None,
            True,
        ),
        (
            sqlite_batch_request,
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT({col}) FROM titanic) FROM {batch} GROUP BY {col}",  # noqa: E501 # FIXME CoP
            False,
            0.2338156892612338,
            'col("Age")<35',
            True,
        ),
        (
            sqlite_batch_request,
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT({col}) FROM {batch}) / 2 FROM {batch} GROUP BY {col}",  # noqa: E501 # FIXME CoP
            False,
            0.3240670220868241,
            None,
            False,
        ),
    ],
)
@pytest.mark.slow  # 3.92s
@pytest.mark.filesystem
def test_expect_queried_column_value_frequency_to_meet_threshold_override_query_sqlite(
    batch_request,
    query,
    success,
    observed,
    row_condition,
    warns,
    titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    validator: Validator = context.get_validator(batch_request=batch_request)

    if warns:
        with pytest.warns(UserWarning):
            result: ExpectationValidationResult = (
                validator.expect_queried_column_value_frequency_to_meet_threshold(
                    column="Sex",
                    value="male",
                    threshold=0.5,
                    query=query,
                    row_condition=row_condition,
                    condition_parser="great_expectations",
                )
            )
    else:
        result: ExpectationValidationResult = (
            validator.expect_queried_column_value_frequency_to_meet_threshold(
                column="Sex",
                value="male",
                threshold=0.5,
                query=query,
                row_condition=row_condition,
                condition_parser="great_expectations",
            )
        )
    assert result["success"] == success and result["result"]["observed_value"] == observed


# noinspection PyUnusedLocal
@pytest.mark.parametrize(
    "success,observed,row_condition,warns",
    [
        (True, 0.6481340441736482, None, False),
        (False, 0.4791666666666667, 'col("Age")<18', True),
        (True, 0.6614626129827444, 'col("Age")>17', True),
    ],
)
@pytest.mark.spark
def test_expect_queried_column_value_frequency_to_meet_threshold_spark(
    success,
    observed,
    row_condition,
    warns,
    spark_session,
    basic_spark_df_execution_engine,
    titanic_df,
):
    df: pd.DataFrame = titanic_df

    context = build_in_memory_runtime_context()

    validator = get_test_validator_with_data(
        execution_engine="spark",
        data=df,
        context=context,
    )

    if warns:
        with pytest.warns(UserWarning):
            result: ExpectationValidationResult = (
                validator.expect_queried_column_value_frequency_to_meet_threshold(
                    column="Sex",
                    value="male",
                    threshold=0.5,
                    row_condition=row_condition,
                    condition_parser="great_expectations",
                )
            )
    else:
        result: ExpectationValidationResult = (
            validator.expect_queried_column_value_frequency_to_meet_threshold(
                column="Sex",
                value="male",
                threshold=0.5,
                row_condition=row_condition,
                condition_parser="great_expectations",
            )
        )
    assert result["success"] == success and result["result"]["observed_value"] == observed


# noinspection PyUnusedLocal
@pytest.mark.parametrize(
    "query,success,observed,row_condition,warns",
    [
        (
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT({col}) FROM {batch}) / 2 FROM {batch} GROUP BY {col}",  # noqa: E501 # FIXME CoP
            False,
            0.3240670220868241,
            None,
            False,
        ),
        (
            "SELECT {col}, CAST(COUNT({col}) AS float) / (SELECT COUNT({col}) FROM {batch}) / 2 FROM {batch} GROUP BY {col}",  # noqa: E501 # FIXME CoP
            False,
            0.3107287449392713,
            'col("Age")<35',
            True,
        ),
    ],
)
@pytest.mark.spark
def test_expect_queried_column_value_frequency_to_meet_threshold_override_query_spark(
    query,
    success,
    observed,
    row_condition,
    warns,
    spark_session,
    basic_spark_df_execution_engine,
    titanic_df,
):
    df: pd.DataFrame = titanic_df

    context = build_in_memory_runtime_context()

    validator = get_test_validator_with_data(
        execution_engine="spark",
        data=df,
        context=context,
    )

    if warns:
        with pytest.warns(UserWarning):
            result: ExpectationValidationResult = (
                validator.expect_queried_column_value_frequency_to_meet_threshold(
                    column="Sex",
                    value="male",
                    threshold=0.5,
                    query=query,
                    row_condition=row_condition,
                    condition_parser="great_expectations",
                )
            )
    else:
        result: ExpectationValidationResult = (
            validator.expect_queried_column_value_frequency_to_meet_threshold(
                column="Sex",
                value="male",
                threshold=0.5,
                query=query,
                row_condition=row_condition,
                condition_parser="great_expectations",
            )
        )
    assert result["success"] == success and result["result"]["observed_value"] == observed


@pytest.mark.big
def test_expect_queried_column_value_frequency_to_meet_threshold_sqlite_multi_value(
    titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    context = titanic_v013_multi_datasource_pandas_and_sqlalchemy_execution_engine_data_context_with_checkpoints_v1_with_empty_store_stats_enabled  # noqa: E501 # FIXME CoP

    validator: Validator = context.get_validator(batch_request=sqlite_batch_request)

    with pytest.warns(UserWarning):
        result: ExpectationValidationResult = (
            validator.expect_queried_column_value_frequency_to_meet_threshold(
                column="Sex",
                value=["male", "female"],
                threshold=[0.6, 0.3],
                row_condition='col("Age")>17',
                condition_parser="great_expectations",
            )
        )

    assert result["success"] is True and result["result"]["observed_value"] == [
        0.6393939393939394,
        0.3606060606060606,
    ]



from typing import Optional
from unittest import mock
from uuid import UUID

import pytest

import great_expectations as gx
from great_expectations.analytics.config import (
    ENV_CONFIG,
    Config,
    get_config,
    update_config,
)
from great_expectations.analytics.events import DataContextInitializedEvent
from great_expectations.data_context.types.base import (
    DataContextConfig,
    InMemoryStoreBackendDefaults,
)
from tests.datasource.fluent._fake_cloud_api import FAKE_USER_ID

TESTING_UUID = UUID("00000000-c000-0000-0000-000000000000")


@pytest.fixture(
    scope="function",
    params=[
        (
            Config(
                organization_id=TESTING_UUID,
                user_id=TESTING_UUID,
                data_context_id=None,
                oss_id=None,
                cloud_mode=False,
            ),
            TESTING_UUID,
            {
                "user_id": TESTING_UUID,
                "organization_id": TESTING_UUID,
                "data_context_id": None,
                "oss_id": None,
                "service": "gx-core",
            },
        ),
        (
            Config(),
            None,
            {"data_context_id": None, "oss_id": None, "service": "gx-core"},
        ),
    ],
)
def analytics_config(request):
    base_config = get_config()
    update_config(request.param[0])
    yield request.param[1], request.param[2]
    update_config(base_config)


@pytest.mark.unit
def test_event_identifiers(analytics_config):
    """Validate base event properties based on the analytics config."""
    distinct_id, base_properties = analytics_config
    event = DataContextInitializedEvent()
    properties = event.properties()
    filtered_base_properties = base_properties.copy()
    # All base properties should be in the event properties
    if "user_id" in filtered_base_properties:
        filtered_base_properties.pop("user_id", None)
        filtered_base_properties.pop("organization_id", None)

    assert filtered_base_properties.items() <= properties.items()

    # Service should be set to gx-core
    assert properties["service"] == "gx-core"
    # The distinct_id should be the user_id if it is set, otherwise the oss_id
    assert event.distinct_id == distinct_id
    if "user_id" in base_properties:
        assert event.distinct_id == base_properties["user_id"]
    else:
        assert event.distinct_id == base_properties["oss_id"]


@pytest.mark.xfail(
    reason="The mode is not always set on instantiation. This is a bug. The test will fail if "
    "run in isolation but may pass if another test has run first."
)
@pytest.mark.unit
def test_ephemeral_context_init(monkeypatch):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", True)  # Enable usage stats

    with (
        mock.patch(
            "great_expectations.data_context.data_context.abstract_data_context.init_analytics"
        ) as mock_init,
        mock.patch("posthog.capture") as mock_submit,
    ):
        _ = gx.get_context(mode="ephemeral")

    mock_init.assert_called_once_with(
        enable=True,
        data_context_id=mock.ANY,
        organization_id=None,
        oss_id=mock.ANY,
        user_id=None,
        user_agent_str=None,
        mode="ephemeral",
    )
    mock_submit.assert_called_once_with(
        mock.ANY,
        "data_context.initialized",
        {
            "data_context_id": mock.ANY,
            "oss_id": mock.ANY,
            "service": "gx-core",
            "gx_version": mock.ANY,
            "user_agent_str": None,
            "mode": "ephemeral",
            "$process_person_profile": False,
        },
        groups={"data_context": mock.ANY},
    )


@pytest.mark.unit
def test_ephemeral_context_init_with_optional_fields(monkeypatch):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", True)  # Enable usage stats

    with mock.patch("posthog.capture") as mock_submit:
        user_agent_str = "test / x.x.x"
        _ = gx.get_context(mode="ephemeral", user_agent_str=user_agent_str)

    mock_submit.assert_called_once_with(
        mock.ANY,
        "data_context.initialized",
        {
            "data_context_id": mock.ANY,
            "oss_id": mock.ANY,
            "service": "gx-core",
            "gx_version": mock.ANY,
            "user_agent_str": user_agent_str,
            "mode": "ephemeral",
            "$process_person_profile": False,
        },
        groups={"data_context": mock.ANY},
    )


@pytest.mark.parametrize("user_agent_str", [None, "test / x.x.x"])
@pytest.mark.cloud
def test_cloud_context_init(
    user_agent_str: Optional[str], cloud_api_fake, cloud_details, monkeypatch
):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", True)  # Enable usage stats

    with (
        mock.patch(
            "great_expectations.data_context.data_context.cloud_data_context.init_analytics"
        ) as mock_init,
        mock.patch("posthog.capture") as mock_submit,
    ):
        _ = gx.get_context(
            cloud_access_token=cloud_details.access_token,
            cloud_organization_id=cloud_details.org_id,
            cloud_base_url=cloud_details.base_url,
            cloud_mode=True,
            user_agent_str=user_agent_str,
        )

    mock_init.assert_called_once_with(
        enable=True,
        user_id=UUID(FAKE_USER_ID),  # Should be consistent with the fake Cloud API
        data_context_id=UUID(cloud_details.org_id),
        organization_id=UUID(cloud_details.org_id),
        oss_id=mock.ANY,
        cloud_mode=True,
        mode="cloud",
        user_agent_str=user_agent_str,
    )
    mock_submit.assert_called_once_with(
        mock.ANY,
        "data_context.initialized",
        {
            "data_context_id": mock.ANY,
            "oss_id": mock.ANY,
            "service": "gx-core",
            "gx_version": mock.ANY,
            "user_agent_str": mock.ANY,
            "mode": mock.ANY,
            "$process_person_profile": False,
        },
        groups={"data_context": mock.ANY},
    )


@pytest.mark.parametrize(
    (
        "environment_variable",
        "constructor_variable",
        "expected_value",
        "user_agent_str",
    ),
    [
        (False, None, False, None),
        (False, False, False, None),
        (False, True, True, None),  # enabling in config overrides environment variable
        (True, None, True, None),
        (True, False, False, None),
        (True, True, True, None),
        (True, True, True, "some user agent string"),
    ],
)
@pytest.mark.unit
def test_analytics_enabled_on_load_unless_disabled(
    environment_variable: bool,
    constructor_variable: Optional[bool],
    expected_value: bool,
    user_agent_str: Optional[str],
    monkeypatch,
):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", environment_variable)
    project_config = DataContextConfig(
        store_backend_defaults=InMemoryStoreBackendDefaults(init_temp_docs_sites=True),
        analytics_enabled=constructor_variable,
    )

    with mock.patch(
        "great_expectations.data_context.data_context.abstract_data_context.init_analytics"
    ) as mock_init:
        gx.get_context(
            mode="ephemeral",
            project_config=project_config,
            user_agent_str=user_agent_str,
        )

    if expected_value is not False:
        mock_init.assert_called_with(
            enable=expected_value,
            data_context_id=mock.ANY,
            organization_id=mock.ANY,
            oss_id=mock.ANY,
            user_id=mock.ANY,
            user_agent_str=user_agent_str,
            mode="ephemeral",
        )
    else:
        mock_init.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("user_agent_str", [None, "some user agent string"])
def test_analytics_enabled_on_load__filesystem(
    user_agent_str: Optional[str],
    tmp_path,
    monkeypatch,
):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", True)

    with mock.patch(
        "great_expectations.data_context.data_context.abstract_data_context.init_analytics"
    ) as mock_init:
        gx.get_context(
            mode="file",
            project_root_dir=tmp_path,
            user_agent_str=user_agent_str,
        )

    mock_init.assert_called_with(
        enable=True,
        data_context_id=mock.ANY,
        organization_id=mock.ANY,
        oss_id=mock.ANY,
        user_id=mock.ANY,
        user_agent_str=user_agent_str,
        mode="file",
    )


@pytest.mark.parametrize("environment_variable", [None, False, True])
@pytest.mark.parametrize("constructor_variable", [None, False, True])
@pytest.mark.parametrize("enable_analytics", [False, True])
@pytest.mark.unit
def test_analytics_enabled_after_setting_explicitly(
    environment_variable: bool,
    constructor_variable: Optional[bool],
    enable_analytics: bool,
    monkeypatch,
):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", environment_variable)
    project_config = DataContextConfig(
        store_backend_defaults=InMemoryStoreBackendDefaults(init_temp_docs_sites=True),
        analytics_enabled=constructor_variable,
    )

    with mock.patch(
        "great_expectations.data_context.data_context.abstract_data_context.init_analytics"
    ) as mock_init:
        context = gx.get_context(
            mode="ephemeral",
            project_config=project_config,
        )

        context.enable_analytics(enable_analytics)

    assert context.config.analytics_enabled == enable_analytics
    if enable_analytics is True:
        mock_init.assert_called_with(
            enable=enable_analytics,
            data_context_id=mock.ANY,
            organization_id=mock.ANY,
            oss_id=mock.ANY,
            user_id=mock.ANY,
            user_agent_str=mock.ANY,
            mode="ephemeral",
        )


@pytest.mark.parametrize("initial_user_agent_str", [None, "old user agent string"])
@pytest.mark.parametrize("new_user_agent_str", [None, "new user agent string"])
@pytest.mark.unit
def test_user_agent_str_after_setting_explicitly(
    initial_user_agent_str: Optional[str],
    new_user_agent_str: Optional[str],
    monkeypatch,
):
    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", True)

    with mock.patch(
        "great_expectations.data_context.data_context.abstract_data_context.init_analytics"
    ) as mock_init:
        context = gx.get_context(
            mode="ephemeral",
            user_agent_str=initial_user_agent_str,
        )

        context.set_user_agent_str(new_user_agent_str)

    mock_init.assert_called_with(
        enable=True,
        data_context_id=mock.ANY,
        organization_id=mock.ANY,
        oss_id=mock.ANY,
        user_id=mock.ANY,
        user_agent_str=new_user_agent_str,
        mode="ephemeral",
    )


@pytest.mark.parametrize("remove_profile", [True, False])
@pytest.mark.unit
def test_remove_profile_setting(remove_profile: bool, monkeypatch):
    # Test that remove_profile flag correctly controls the $process_person_profile property.
    # https://posthog.com/docs/libraries/python#person-profiles-and-properties

    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", True)  # Enable usage stats

    with mock.patch("posthog.capture") as mock_submit:
        from great_expectations.analytics.client import init, submit
        from great_expectations.analytics.events import DataContextInitializedEvent

        init(
            enable=True,
            mode="ephemeral",
            data_context_id=TESTING_UUID,
            remove_profile=remove_profile,
        )
        event = DataContextInitializedEvent()

        submit(event)

        mock_submit.assert_called_once()

        args = mock_submit.call_args.args
        properties = args[2]

        if remove_profile:
            assert properties.get("$process_person_profile") is False
        else:
            assert properties.get("$process_person_profile") is True



from __future__ import annotations

from copy import copy
from pprint import pformat as pf
from unittest import mock

import pytest

import great_expectations.expectations as gxe
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.partitioners import PartitionerColumnValue
from great_expectations.core.result_format import ResultFormat
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.datasource.fluent.interfaces import DataAsset, Datasource
from great_expectations.expectations.expectation import Expectation
from great_expectations.validator.v1_validator import Validator


@pytest.fixture
def failing_expectation() -> Expectation:
    return gxe.ExpectColumnValuesToBeInSet(
        column="event_type",
        value_set=["start", "stop"],
    )


@pytest.fixture
def passing_expectation() -> Expectation:
    return gxe.ExpectColumnValuesToBeBetween(
        column="id",
        min_value=-1,
        max_value=1000000,
    )


@pytest.fixture
def expectation_suite(
    failing_expectation: Expectation, passing_expectation: Expectation
) -> ExpectationSuite:
    suite = ExpectationSuite("test_suite")
    suite.add_expectation_configuration(failing_expectation.configuration)
    suite.add_expectation_configuration(passing_expectation.configuration)
    return suite


@pytest.fixture
def fds_data_asset(
    fds_data_context: AbstractDataContext,
    fds_data_context_datasource_name: str,
) -> DataAsset:
    datasource = fds_data_context.data_sources.get(fds_data_context_datasource_name)
    assert isinstance(datasource, Datasource)
    return datasource.get_asset("trip_asset")


@pytest.fixture
def fds_data_asset_with_event_type_partitioner(
    fds_data_context: AbstractDataContext,
    fds_data_context_datasource_name: str,
) -> DataAsset:
    datasource = fds_data_context.data_sources.get(fds_data_context_datasource_name)
    assert isinstance(datasource, Datasource)
    return datasource.get_asset("trip_asset_partition_by_event_type")


@pytest.fixture
def batch_definition(
    fds_data_asset: DataAsset,
) -> BatchDefinition:
    batch_definition = BatchDefinition[None](name="test_batch_definition")
    batch_definition.set_data_asset(fds_data_asset)
    return batch_definition


@pytest.fixture
def batch_definition_with_event_type_partitioner(
    fds_data_asset_with_event_type_partitioner: DataAsset,
) -> BatchDefinition:
    partitioner = PartitionerColumnValue(column_name="event_type")
    batch_definition = BatchDefinition(name="test_batch_definition", partitioner=partitioner)
    batch_definition.set_data_asset(fds_data_asset_with_event_type_partitioner)
    return batch_definition


@pytest.fixture
def validator(
    fds_data_context: AbstractDataContext, batch_definition: BatchDefinition
) -> Validator:
    return Validator(
        batch_definition=batch_definition,
        batch_parameters=None,
        result_format=ResultFormat.SUMMARY,
    )


@pytest.mark.unit
def test_result_format_boolean_only(validator: Validator, failing_expectation: Expectation):
    validator.result_format = ResultFormat.BOOLEAN_ONLY
    result = validator.validate_expectation(failing_expectation)

    assert not result.success
    assert result.result == {}


@pytest.mark.unit
def test_result_format_basic(validator: Validator, failing_expectation: Expectation):
    validator.result_format = ResultFormat.BASIC
    result = validator.validate_expectation(failing_expectation)

    assert not result.success

    assert "partial_unexpected_list" in result.result
    assert "partial_unexpected_counts" not in result.result
    assert "unexpected_list" not in result.result


@pytest.mark.unit
def test_result_format_summary(validator: Validator, failing_expectation: Expectation):
    validator.result_format = ResultFormat.SUMMARY
    result = validator.validate_expectation(failing_expectation)

    assert not result.success

    assert "partial_unexpected_list" in result.result
    assert "partial_unexpected_counts" in result.result
    assert "unexpected_list" not in result.result


@pytest.mark.unit
def test_result_format_complete(validator: Validator, failing_expectation: Expectation):
    validator.result_format = ResultFormat.COMPLETE
    result = validator.validate_expectation(failing_expectation)

    assert not result.success

    assert "partial_unexpected_list" in result.result
    assert "partial_unexpected_counts" in result.result
    assert "unexpected_list" in result.result


@pytest.mark.unit
def test_v1_validator_doesnt_mutate_result_format(
    validator: Validator, expectation_suite: ExpectationSuite
):
    """This test verifies a bugfix where the legacy Validator mutates a ResultFormat
    dict provided by the user.
    """
    result_format_dict = {
        "result_format": "COMPLETE",
    }
    backup_result_format_dict = copy(result_format_dict)
    validator.result_format = result_format_dict
    validator.validate_expectation_suite(expectation_suite=expectation_suite)
    assert result_format_dict == backup_result_format_dict


@pytest.mark.unit
def test_validate_expectation_success(validator: Validator, passing_expectation: Expectation):
    result = validator.validate_expectation(passing_expectation)

    assert result.success


@pytest.mark.unit
def test_validate_expectation_failure(validator: Validator, failing_expectation: Expectation):
    result = validator.validate_expectation(failing_expectation)

    assert not result.success


@pytest.mark.unit
def test_validate_expectation_with_batch_asset_options(
    fds_data_context: AbstractDataContext,
    batch_definition_with_event_type_partitioner: BatchDefinition,
):
    desired_event_type = "start"
    validator = Validator(
        batch_definition=batch_definition_with_event_type_partitioner,
        batch_parameters={"event_type": desired_event_type},
    )

    result = validator.validate_expectation(
        gxe.ExpectColumnValuesToBeInSet(
            column="event_type",
            value_set=[desired_event_type],
        )
    )
    print(f"Result dict ->\n{pf(result)}")
    assert result.success


@pytest.mark.unit
def test_validate_expectation_suite(validator: Validator, expectation_suite: ExpectationSuite):
    result = validator.validate_expectation_suite(expectation_suite)

    assert not result.success
    assert not result.results[0].success
    assert result.results[1].success
    assert result.statistics == {
        "evaluated_expectations": 2,
        "successful_expectations": 1,
        "unsuccessful_expectations": 1,
        "success_percent": 50.0,
    }


@pytest.mark.parametrize(
    ["parameter", "expected"],
    [
        (["start", "stop", "continue"], True),
        (["start", "stop"], False),
    ],
)
@pytest.mark.unit
def test_validate_expectation_suite_suite_parameters(
    validator: Validator,
    parameter: list[str],
    expected: bool,
):
    suite = ExpectationSuite("test_suite")
    expectation = gxe.ExpectColumnValuesToBeInSet(
        column="event_type",
        value_set={"$PARAMETER": "my_parameter"},
    )
    suite.add_expectation_configuration(expectation.configuration)
    result = validator.validate_expectation_suite(suite, {"my_parameter": parameter})

    assert result.success == expected


@pytest.mark.unit
def test_non_cloud_validate_does_not_render_results(
    validator: Validator,
    empty_data_context: AbstractDataContext,
):
    suite = empty_data_context.suites.add(
        ExpectationSuite(
            name="test_suite",
            expectations=[
                gxe.ExpectColumnValuesToBeInSet(
                    column="event_type",
                    value_set=["start"],
                )
            ],
        )
    )
    result = validator.validate_expectation_suite(suite)

    assert len(result.results) == 1
    assert not result.results[0].rendered_content


@mock.patch(
    "great_expectations.data_context.data_context.context_factory.project_manager.is_using_cloud",
)
@pytest.mark.unit
def test_cloud_validate_renders_results_when_appropriate(
    mock_is_using_cloud,
    validator: Validator,
    empty_data_context: AbstractDataContext,
):
    mock_is_using_cloud.return_value = True
    suite = empty_data_context.suites.add(
        ExpectationSuite(
            name="test_suite",
            expectations=[
                gxe.ExpectColumnValuesToBeInSet(
                    column="event_type",
                    value_set=["start"],
                )
            ],
        )
    )
    result = validator.validate_expectation_suite(suite)

    assert len(result.results) == 1
    assert result.results[0].rendered_content



import pytest

from great_expectations.core.expectation_validation_result import (
    ExpectationValidationResult,
)
from great_expectations.validator.validation_statistics import (
    ValidationStatistics,
    calc_validation_statistics,
)


@pytest.mark.unit
def test_stats_no_expectations():
    expectation_results = []
    actual = calc_validation_statistics(expectation_results)

    # pay attention to these two
    assert None is actual.success_percent
    assert True is actual.success
    # the rest is boring
    assert actual.successful_expectations == 0
    assert actual.evaluated_expectations == 0
    assert actual.unsuccessful_expectations == 0


@pytest.mark.unit
def test_stats_no_successful_expectations():
    expectation_results = [ExpectationValidationResult(success=False)]
    actual = calc_validation_statistics(expectation_results)
    expected = ValidationStatistics(1, 0, 1, 0.0, False)
    assert expected == actual

    expectation_results = [
        ExpectationValidationResult(success=False),
        ExpectationValidationResult(success=False),
        ExpectationValidationResult(success=False),
    ]
    actual = calc_validation_statistics(expectation_results)
    expected = ValidationStatistics(3, 0, 3, 0.0, False)
    assert expected == actual


@pytest.mark.unit
def test_stats_all_successful_expectations():
    expectation_results = [
        ExpectationValidationResult(success=True),
    ]
    actual = calc_validation_statistics(expectation_results)
    expected = ValidationStatistics(1, 1, 0, 100.0, True)
    assert expected == actual

    expectation_results = [
        ExpectationValidationResult(success=True),
        ExpectationValidationResult(success=True),
        ExpectationValidationResult(success=True),
    ]
    actual = calc_validation_statistics(expectation_results)
    expected = ValidationStatistics(3, 3, 0, 100.0, True)
    assert expected == actual


@pytest.mark.unit
def test_stats_mixed_expectations():
    expectation_results = [
        ExpectationValidationResult(success=False),
        ExpectationValidationResult(success=True),
    ]
    actual = calc_validation_statistics(expectation_results)
    expected = ValidationStatistics(2, 1, 1, 50.0, False)
    assert expected == actual
